"""Workflow tasks."""
import luigi
from data_validation_framework.task import ValidationWorkflow

from morphology_workflows.tasks.annotation import ApicalPoint
from morphology_workflows.tasks.annotation import CollectCurated
from morphology_workflows.tasks.annotation import CutLeaves
from morphology_workflows.tasks.annotation import HardLimit
from morphology_workflows.tasks.annotation import MType
from morphology_workflows.tasks.annotation import PlotApicalPoint
from morphology_workflows.tasks.annotation import PlotCutLeaves
from morphology_workflows.tasks.annotation import PlotHardLimit
from morphology_workflows.tasks.curation import Align
from morphology_workflows.tasks.curation import CheckNeurites
from morphology_workflows.tasks.curation import Collect
from morphology_workflows.tasks.curation import DetectErrors
from morphology_workflows.tasks.curation import EnsureNeuritesOutsideSoma
from morphology_workflows.tasks.curation import ErrorsReport
from morphology_workflows.tasks.curation import ExtractMarkers
from morphology_workflows.tasks.curation import Orient
from morphology_workflows.tasks.curation import PlotErrors
from morphology_workflows.tasks.curation import PlotMarkers
from morphology_workflows.tasks.curation import PlotMorphologies
from morphology_workflows.tasks.curation import Recenter
from morphology_workflows.tasks.curation import Resample
from morphology_workflows.tasks.curation import Sanitize
from morphology_workflows.tasks.repair import CollectAnnotated
from morphology_workflows.tasks.repair import FixZeroDiameters
from morphology_workflows.tasks.repair import MakeCollage
from morphology_workflows.tasks.repair import MakeRelease
from morphology_workflows.tasks.repair import PlotRepair
from morphology_workflows.tasks.repair import PlotSmoothDiameters
from morphology_workflows.tasks.repair import RepairNeurites
from morphology_workflows.tasks.repair import SmoothDiameters
from morphology_workflows.tasks.repair import Unravel


def save_reduced_df(
    df, data_dir, df_path="reduced_df.csv", parents=None
):  # pylint: disable=unused-argument
    """Save reduced version of the report."""
    _to_remove = ["exception", "is_valid", "ret_code", "comment"]
    if parents:
        out_dir = list(data_dir.parents)[parents - 1]
    else:
        out_dir = data_dir
    df.loc[
        df.is_valid, [col for col in df.columns if isinstance(col, str) and col not in _to_remove]
    ].rename_axis(index="morph_name").reset_index().to_csv(out_dir / df_path, index=False)


class Curate(ValidationWorkflow):
    """Run Curation phase.

    The complete phase has the following dependency graph:

    .. graphviz:: Curate.dot
    """

    __specifications__ = "Run the Curation phase."

    input_index_col = luigi.Parameter(default="morph_name")
    args = ["curated_dataset.csv", 2]

    validation_function = staticmethod(save_reduced_df)

    def inputs(self):
        """ """
        return {
            Collect: {},
            ExtractMarkers: {"marker_path": "marker_path"},
            Sanitize: {},
            CheckNeurites: {
                "has_axon": "has_axon",
                "has_basal": "has_basal",
                "has_apical": "has_apical",
            },
            Recenter: {"soma_location": "soma_location"},
            DetectErrors: {
                "error_marker_path": "error_marker_path",
                "error_annotated_path": "error_annotated_path",
            },
            PlotMarkers: {},
            PlotErrors: {},
            ErrorsReport: {},
            Align: {"rotation_matrix": "rotation_matrix"},
            EnsureNeuritesOutsideSoma: {},
            Orient: {},
            Resample: {"morph_path": "morph_path"},
            PlotMorphologies: {},
        }


class Annotate(ValidationWorkflow):
    """Run Annotation phase.

    The complete phase has the following dependency graph:

    .. graphviz:: Annotate.dot
    """

    __specifications__ = "Run the Annotation phase."

    input_index_col = luigi.Parameter(default="morph_name")
    args = ["annotated_dataset.csv", 2]

    validation_function = staticmethod(save_reduced_df)

    def inputs(self):
        """ """
        return {
            CollectCurated: {},
            MType: {"mtype": "mtype"},
            HardLimit: {"hard_limit_path": "hard_limit_path"},
            ApicalPoint: {"apical_point_path": "apical_point_path"},
            CutLeaves: {"cut_leaves_path": "cut_leaves_path"},
            PlotApicalPoint: {},
            PlotCutLeaves: {},
            PlotHardLimit: {},
        }


class Repair(ValidationWorkflow):
    """Run Repair phase.

    The complete phase has the following dependency graph:

    .. graphviz:: Repair.dot
    """

    __specifications__ = "Run the Repair phase."

    input_index_col = luigi.Parameter(default="morph_name")
    make_release = luigi.BoolParameter(
        default=True, description=":bool: Set to True to make a morpology release with neurondb.xml"
    )
    args = ["repaired_dataset.csv", 2]

    validation_function = staticmethod(save_reduced_df)

    report_config = {
        "extensions": [
            "sphinx.ext.graphviz",
            "sphinx.ext.intersphinx",
            "sphinx.ext.napoleon",
            "sphinx.ext.todo",
            "rst2pdf.pdfbuilder",
        ]
    }

    def inputs(self):
        """ """
        inputs = {
            CollectAnnotated: {
                "morph_path": "morph_path",
                "apical_point_path": "apical_point_path",
                "cut_leaves_path": "cut_leaves_path",
            },
            FixZeroDiameters: {"morph_path": "zero_diameter_morph_path"},
            Unravel: {"morph_path": "unravel_morph_path"},
            RepairNeurites: {"morph_path": "repair_morph_path"},
            SmoothDiameters: {"morph_path": "smooth_morph_path"},
            PlotSmoothDiameters: {},
            PlotRepair: {},
            MakeCollage: {},
        }

        if self.make_release:
            folders = [
                "zero_diameter_morph_db_path",
                "unravel_morph_db_path",
                "repair_morph_db_path",
                "zero_diameter_release_morph_path",
                "unravel_release_morph_path",
                "repair_release_morph_path",
            ]
            inputs[MakeRelease] = {}
            for extension in MakeRelease().extensions:
                ext = extension[1:]
                inputs[MakeRelease].update(
                    {f"{folder}_{ext}": f"{folder}_{ext}" for folder in folders}
                )

        return inputs



"""Tests suite for the morphology-workflows package."""
import re
from tempfile import NamedTemporaryFile

from neurom.core import Morphology


def remove_between(start_part, end_part, tags=None):
    """Prepare a regex to replace a part of a string by three dots."""
    if tags:
        return (f"{start_part}.*{end_part}", f"{start_part} ... {end_part}", tags)
    else:
        return (f"{start_part}.*{end_part}", f"{start_part} ... {end_part}")


def clean_exception(exc):
    """Prepare a regex to clean an exception."""
    return remove_between("Traceback ", exc, re.DOTALL)


def create_morphology(content, extension):
    """Create a morphology from a string."""
    with NamedTemporaryFile(suffix=f".{extension}", mode="w", encoding="utf-8") as tmp_file:
        tmp_file.write(content)
        tmp_file.flush()
        return Morphology(tmp_file.name)


def create_morphology_file(content, extension, filepath):
    """Create a morphology file from a string."""
    morph = create_morphology(content, extension)
    morph.write(filepath)
    return filepath



"""Process functions."""
import logging
import shutil
from functools import partial
from multiprocessing.pool import Pool
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from data_validation_framework.result import ValidationResult
from diameter_synthesis.main import diametrize_single_neuron
from matplotlib.backends.backend_pdf import PdfPages
from morph_tool.converter import convert
from morph_tool.exceptions import MorphToolException
from morph_tool.morphdb import MorphDB
from morph_tool.morphdb import MorphInfo
from morphio.mut import Morphology
from neurom import load_morphology
from neurom import view
from neuror.main import RepairType
from neuror.main import repair as _repair
from neuror.unravel import unravel as _unravel
from neuror.zero_diameter_fixer import fix_zero_diameters as _fix_zero_diameters
from scipy.spatial import KDTree
from tqdm import tqdm

from morphology_workflows.marker_helper import MarkerSet
from morphology_workflows.utils import disable_loggers

L = logging.getLogger(__name__)
matplotlib.use("Agg")


def fix_zero_diameters(row, data_dir):
    """Assign mtype to morphologies."""
    neuron = Morphology(row.morph_path)
    _fix_zero_diameters(neuron)
    morph_path = data_dir / Path(row.morph_path).name
    neuron.write(morph_path)
    return ValidationResult(is_valid=True, morph_path=morph_path)


def _map_apical(apical_point, mapping):
    """Map apical points as cut-leaves."""
    t = KDTree(mapping[["x0", "y0", "z0"]])
    _, indices = t.query(apical_point)
    return mapping.iloc[indices][["x1", "y1", "z1"]].values


def _unravel_leaves(leaves, mapping):
    """Adapted from neuror.unraavel.unravel_plane."""
    t = KDTree(mapping[["x0", "y0", "z0"]])
    distances, indices = t.query(leaves)
    not_matching_leaves = np.where(distances > 1e-3)[0]
    if not_matching_leaves.size:
        raise Exception(
            f"Cannot find the following leaves in the mapping:\n{leaves[not_matching_leaves]}"
        )
    return mapping.iloc[indices][["x1", "y1", "z1"]].values


def unravel(row, data_dir, window_half_length=10):
    """Unravel morphologies and update cut plane."""
    neuron, mapping = _unravel(row.morph_path, window_half_length=window_half_length)
    morph_path = data_dir / Path(row.morph_path).name
    neuron.write(morph_path)

    # transform original cut leaves
    if not row.isnull()["cut_leaves_path"]:
        cut_leaves_marker = MarkerSet.from_file(row.cut_leaves_path)
        cut_leaves_marker.markers[0].data = _unravel_leaves(
            cut_leaves_marker.markers[0].data, mapping
        )
        new_cut_leaves_path = data_dir / (row.name + ".yaml")
        cut_leaves_marker.save(filename=new_cut_leaves_path)

        # transform original apical points
        new_apical_point_path = np.nan
        if not row.isnull()["apical_point_path"]:
            new_apical_point_path = (data_dir / ("apical_" + row.name)).with_suffix(".yaml")
            apical_point = MarkerSet.from_file(row.apical_point_path)
            apical_point.markers[0].data = _map_apical(
                np.array(apical_point.markers[0].data), mapping
            )
            apical_point.save(filename=new_apical_point_path)
    else:
        new_cut_leaves_path = None
        new_apical_point_path = None

    return ValidationResult(
        is_valid=True,
        morph_path=morph_path,
        unravelled_cut_leaves_path=new_cut_leaves_path,
        unravelled_apical_point_path=new_apical_point_path,
    )


def repair(row, data_dir, with_plot=False, repair_params=None):
    """Repair morphologies.

    TODO: understand the repair_flag correctly and how to set them.
    """
    morph_path = data_dir / Path(row.morph_path).name
    apical_point = -1
    if not row.isnull()["unravelled_apical_point_path"]:
        apical_point = (
            MarkerSet.from_file(row.unravelled_apical_point_path).markers[0].data.tolist()
        )

    if not row.isnull()["unravelled_cut_leaves_path"]:
        leaves = MarkerSet.from_file(row.unravelled_cut_leaves_path).markers[0].data
        _repair(  # pylint: disable=unexpected-keyword-arg
            row.morph_path,
            morph_path,
            axons=None,
            seed=0,
            cut_leaves_coordinates=leaves,
            legacy_detection=False,  # cut_leaves_coordinates are provided, so this is not used
            plot_file=(data_dir / row.name).with_suffix(".html") if with_plot else None,
            repair_flags={
                RepairType.axon: row.has_axon,
                RepairType.tuft: row.has_apical,
                RepairType.oblique: row.has_apical,
                RepairType.basal: row.has_basal,
            },
            apical_point=apical_point,
            params=repair_params,
        )
    else:
        shutil.copy(row.morph_path, morph_path)
    return ValidationResult(is_valid=True, morph_path=morph_path)


def plot_repair(row, data_dir, with_plotly=True):
    """Plotting cut leaves on morphologies."""
    markers = None
    if not row.isnull()["cut_leaves_path"]:
        markers = MarkerSet.from_file(row.cut_leaves_path)
    plot_path = None
    if with_plotly:
        if markers is not None:
            plot_path = (data_dir / row.name).with_suffix(".html")
            markers.plot(filename=plot_path)

    else:
        matplotlib.font_manager._get_font.cache_clear()  # pylint: disable=protected-access
        plot_path = (data_dir / row.name).with_suffix(".pdf")
        neuron = load_morphology(row.morph_path)

        points = None
        if markers is not None:
            points = markers.markers[0].data.T[[0, 2]]

        with disable_loggers(
            "matplotlib.font_manager", "matplotlib.backends.backend_pdf"
        ), PdfPages(plot_path) as pdf:
            for plane in ["xy", "xz", "yz"]:
                plt.figure()
                ax = plt.gca()

                view.plot_morph(
                    neuron, ax, realistic_diameters=True, plane="xz", soma_outline=False
                )
                if points is not None:
                    plt.scatter(*points, color="g", s=2, marker="o")
                ax.autoscale()
                ax.axis("equal")
                plt.title(f"{row.name}")
                if hasattr(row, "mtype") and isinstance(row.mtype, str):
                    plt.suptitle(plane + "mtype:" + row.mtype)
                else:
                    plt.suptitle(plane)
                pdf.savefig()
                plt.close()

    return ValidationResult(is_valid=True, plot_repair_path=plot_path)


def smooth_diameters(row, data_dir):
    """Smooth diameters using diameter-synthesis simpler algorithm."""
    morph = Morphology(row.morph_path)
    diametrize_single_neuron(morph)
    morph_path = data_dir / Path(row.morph_path).name
    morph.write(morph_path)
    return ValidationResult(is_valid=True, morph_path=morph_path)


def plot_smooth_diameters(row, data_dir, shift=200):
    """Plot original morphology and smoother one next to each other."""
    orig_neuron = load_morphology(row.morph_path)
    smooth_neuron = load_morphology(row.smooth_morph_path)
    plt.figure()
    ax = plt.gca()
    view.plot_morph(
        orig_neuron.transform(lambda p: p - np.array([shift, 0, 0])),
        ax=ax,
        plane="xy",
        soma_outline=False,
        realistic_diameters=False,
    )
    view.plot_morph(
        smooth_neuron.transform(lambda p: p + np.array([shift, 0, 0])),
        ax=ax,
        plane="xy",
        soma_outline=False,
        realistic_diameters=False,
    )
    plot_path = (data_dir / row.name).with_suffix(".pdf")
    plt.autoscale()
    plt.axis("equal")
    plt.savefig(plot_path, bbox_inches="tight")
    return ValidationResult(is_valid=True, plot_smooth_path=plot_path)


# pylint: disable=too-many-arguments,too-many-locals
def make_collage(
    df,
    data_dir,
    collage_path="collage.pdf",
    separation=500,
    n_morph_per_page=10,
    middle_panel_shift=1000,
    top_panel_shift=2000,
    layer_thickness=None,
    rasterized=False,
    dpi=1000,
    figsize=None,
):
    """Make collage plot of morphologies per mtype."""
    if layer_thickness is None:
        layer_thickness = [700, 525, 190, 353, 149, 165]

    if figsize is None:
        figsize = (12, 10)

    layer_boundaries = np.insert(np.cumsum(layer_thickness), 0, 0)
    layer_centers = 0.5 * (layer_boundaries[:-1] + layer_boundaries[1:])[::-1]
    middle_panel_shift += layer_boundaries[-1]
    top_panel_shift += layer_boundaries[-1]

    mtypes = sorted(df.mtype.unique())
    with disable_loggers("matplotlib.font_manager", "matplotlib.backends.backend_pdf"), PdfPages(
        data_dir.parent.parent / collage_path
    ) as pdf:
        for mtype in tqdm(mtypes):
            _df = df[df.mtype == mtype]
            name_batches = np.array_split(_df.index, max(1, len(_df.index) / n_morph_per_page))
            for page, batch in enumerate(name_batches):
                plt.figure(figsize=figsize)
                ax = plt.gca()
                for shift, morph_name in enumerate(batch):
                    try:
                        layer_pos = layer_centers[int(mtype[1]) - 1]
                    except (TypeError, IndexError):
                        layer_pos = layer_centers.mean()
                    neuron = load_morphology(_df.loc[morph_name, "morph_path"])

                    translate = [shift * separation, layer_pos, 0.0]
                    _neuron = neuron.transform(lambda p, t=translate: p + t)
                    view.plot_morph(
                        _neuron,
                        ax,
                        plane="xy",
                        soma_outline=False,
                        realistic_diameters=False,
                    )
                    ax.axhline(middle_panel_shift, c="k", lw=0.5, ls="--")

                    translate = [shift * separation, 0, middle_panel_shift]
                    _neuron = neuron.transform(lambda p, t=translate: p + t)
                    view.plot_morph(
                        _neuron,
                        ax,
                        plane="xz",
                        soma_outline=False,
                        realistic_diameters=False,
                    )
                    ax.axhline(top_panel_shift, c="k", lw=0.5, ls="--")

                    translate = [0, shift * separation, top_panel_shift]
                    _neuron = neuron.transform(lambda p, t=translate: p + t)
                    view.plot_morph(
                        _neuron,
                        ax,
                        plane="yz",
                        soma_outline=False,
                        realistic_diameters=False,
                    )
                    ax.text(
                        shift * separation - separation / 2.0,
                        layer_boundaries[0] - 100,
                        morph_name,
                        fontsize="x-small",
                        rotation=45,
                    )

                for lb in layer_boundaries:
                    ax.axhline(lb, color="r", ls="--", alpha=0.3)

                ax.set_rasterized(rasterized)
                ax.set_title(f"{mtype}, page {page + 1} / {len(name_batches)}")
                ax.set_ylim(layer_boundaries[0] - 150, top_panel_shift + 100)
                ax.set_aspect("equal")
                pdf.savefig(bbox_inches="tight", dpi=dpi)
                plt.close()


def _convert(input_file, output_file):
    """Handles crashes in conversion of writing of morphologies."""
    try:
        convert(input_file, output_file)
        return output_file
    except MorphToolException:
        return "cannot save"
    except RuntimeError:  # this can happen if duplicates are being written as the same time
        return output_file


def _create_db_row(_data, zero_diameter_path, unravel_path, repair_path, extension):
    """Create a db row and convert morphology."""
    index, data = _data
    m = MorphInfo(
        name=data["morph_name"],
        mtype=data["mtype"],
        layer=int(data["layer"]),
        use_dendrite=data["has_basal"],
        use_axon=data["has_axon"],
    )

    if zero_diameter_path is not None:
        zero_diameter_release_path = (
            str(zero_diameter_path / Path(data["zero_diameter_morph_path"]).stem) + extension
        )

        data[f"zero_diameter_release_morph_path_{extension[1:]}"] = _convert(
            data["zero_diameter_morph_path"], zero_diameter_release_path
        )

    if unravel_path is not None:
        unravel_release_path = str(unravel_path / Path(data["unravel_morph_path"]).stem) + extension
        data[f"unravel_release_morph_path_{extension[1:]}"] = _convert(
            data["unravel_morph_path"], unravel_release_path
        )

    if repair_path is not None:
        repair_release_path = str(repair_path / Path(data["repair_morph_path"]).stem) + extension
        data[f"repair_release_morph_path_{extension[1:]}"] = _convert(
            data["repair_morph_path"], repair_release_path
        )
    return index, data, m


def set_layer_column(df):
    """Set layer values from mtype name if no layer column is present."""
    for gid in df.index:
        if df.loc[gid, "layer"] is None:
            mtype = df.loc[gid, "mtype"]
            if isinstance(mtype, str):
                if len(mtype) > 1:
                    try:
                        layer = int(mtype[1])
                    except ValueError:
                        layer = 0
            else:
                layer = 0
            df.loc[gid, "layer"] = layer
            df.loc[gid, "mtype"] = mtype


def add_duplicated_layers(df):
    """Duplicate entries if layer name has mixed layers, i.e. L23_PC."""
    for mtype in df.mtype.unique():
        layer = mtype.split("_")[0][1:]
        if len(layer) > 1:
            _df = df[df.mtype == mtype]
            _df.layer = int(layer[1])
            df = pd.concat([df, _df]).reset_index(drop=True)
    return df


def make_release(
    df, _, zero_diameter_path, unravel_path, repair_path, extensions, duplicate_layers=True
):
    """Make morphology release."""
    set_layer_column(df)

    df_tmp = df.reset_index()
    if duplicate_layers:
        df_tmp = add_duplicated_layers(df_tmp)

    for extension in extensions:
        _zero_diameter_path = None
        if zero_diameter_path is not None:
            _zero_diameter_path = Path(f"{zero_diameter_path}-{extension[1:]}")
            _zero_diameter_path.mkdir(exist_ok=True, parents=True)

        _unravel_path = None
        if unravel_path is not None:
            _unravel_path = Path(f"{unravel_path}-{extension[1:]}")
            _unravel_path.mkdir(exist_ok=True, parents=True)

        _repair_path = None
        if repair_path is not None:
            _repair_path = Path(f"{repair_path}-{extension[1:]}")
            _repair_path.mkdir(exist_ok=True, parents=True)

        __create_db_row = partial(
            _create_db_row,
            zero_diameter_path=_zero_diameter_path,
            unravel_path=_unravel_path,
            repair_path=_repair_path,
            extension=extension,
        )

        _m = []
        written = set()
        with Pool() as pool:
            for index, row, m in tqdm(
                pool.imap(__create_db_row, df_tmp.loc[df_tmp["is_valid"]].iterrows()),
                total=len(df_tmp),
            ):
                if row["morph_name"] in df.index and row["morph_name"] not in written:
                    df.loc[row["morph_name"]] = pd.Series(row)
                    written.add(row["morph_name"])
                df_tmp.loc[index] = pd.Series(row)
                _m.append(m)

        db = MorphDB(_m)
        if _zero_diameter_path is not None:
            db.write(_zero_diameter_path / "neurondb.xml")
            col_name = f"zero_diameter_morph_db_path_{extension[:1]}"
            db_path = _zero_diameter_path / "neurondb.xml"
            df.loc[df["is_valid"], col_name] = db_path
            df_tmp.loc[df_tmp["is_valid"], col_name] = db_path

        if _unravel_path is not None:
            db.write(_unravel_path / "neurondb.xml")
            col_name = f"unravel_morph_db_path_{extension[1:]}"
            db_path = _unravel_path / "neurondb.xml"
            df.loc[df["is_valid"], col_name] = db_path
            df_tmp.loc[df_tmp["is_valid"], col_name] = db_path

        if _repair_path is not None:
            db.write(_repair_path / "neurondb.xml")
            col_name = f"repair_morph_db_path_{extension[1:]}"
            db_path = _repair_path / "neurondb.xml"
            df.loc[df["is_valid"], col_name] = db_path
            df_tmp.loc[df_tmp["is_valid"], col_name] = db_path



"""Configuration for the pytest test suite."""
import os
from pathlib import Path

import dir_content_diff
import dir_content_diff.pandas
import pytest

DATA = Path(__file__).parent / "data"
EXAMPLES = Path(__file__).parent / "examples_test"

dir_content_diff.pandas.register()


@pytest.fixture()
def tmp_working_dir(tmp_path):
    """Change working directory before a test and change it back when the test is finished."""
    cwd = os.getcwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(cwd)


@pytest.fixture()
def data_dir():
    """The data directory."""
    return DATA


@pytest.fixture()
def examples_dir():
    """The examples directory."""
    return EXAMPLES



"""Util functions."""
import logging
from contextlib import contextmanager
from functools import wraps
from pathlib import Path

from morphio.mut import Morphology
from tqdm import tqdm

tqdm.pandas()

EXTS = {".asc", ".h5", ".swc"}  # allowed extensions


def is_morphology(filename):
    """Returns True if the extension is supported."""
    try:
        Morphology(filename)
        ext = Path(filename).suffix.lower()
        return ext in EXTS, ext
    except Exception:  # pylint: disable=broad-except
        return False, None


def silent_logger(log_name):
    """A decorator to silent a logger during the function execution."""

    def _silent_logger(function):
        @wraps(function)
        def decorated_func(*args, **kwargs):
            func_logger = logging.getLogger(log_name)
            func_logger.disabled = True
            try:
                return function(*args, **kwargs)
            finally:
                func_logger.disabled = False

        return decorated_func

    return _silent_logger


@contextmanager
def disable_loggers(*logger_names):
    """A context manager to silent loggers during the body execution.

    Args:
        *logger_names (str): The names of the loggers to be disabled.
    """
    if not logger_names:
        loggers = [logging.root]
    else:
        loggers = [logging.getLogger(i) for i in logger_names]

    disabled_loggers = [(i, i.disabled) for i in loggers]

    try:
        for i, _ in disabled_loggers:
            i.disabled = True
        yield
    finally:
        for i, j in disabled_loggers:
            i.disabled = j



"""Repair tasks."""
import logging

import luigi
from data_validation_framework.task import ElementValidationTask
from data_validation_framework.task import SetValidationTask
from data_validation_framework.task import SkippableMixin
from luigi_tools.parameter import BoolParameter

from morphology_workflows.curation import collect
from morphology_workflows.repair import fix_zero_diameters
from morphology_workflows.repair import make_collage
from morphology_workflows.repair import make_release
from morphology_workflows.repair import plot_repair
from morphology_workflows.repair import plot_smooth_diameters
from morphology_workflows.repair import repair
from morphology_workflows.repair import smooth_diameters
from morphology_workflows.repair import unravel

logger = logging.getLogger(__name__)


class CollectAnnotated(ElementValidationTask):
    """Collect annotated dataset to work with on this phase."""

    output_columns = {
        "morph_path": None,
        "has_apical": None,
        "has_axon": None,
        "has_basal": None,
        "cut_leaves_path": None,
        "apical_point_path": None,
    }

    validation_function = collect


class FixZeroDiameters(ElementValidationTask):
    """Fix zero diameters.

    This task applies a fix on zero diameters on dendrites, by calling
    :func:`neuror.zero_diameter_fixer.fix_zero_diameters`.
    """

    output_columns = {"morph_path": None}

    validation_function = fix_zero_diameters

    def inputs(self):
        """ """
        return {CollectAnnotated: {"morph_path": "morph_path"}}


class Unravel(ElementValidationTask):
    """Unravel morphologies.

    In-vitro morphologies produce recostruction with too much tortuosity, which is corrected for
    here, using :func:`neuror.unravel.unravel`.
    As this changes the location of the points, the cut leaves have to be updated, as well as the
    apical points, if any.
    """

    window_half_length = luigi.FloatParameter(
        default=10, description=":float: Size in path length of unravel window"
    )
    output_columns = {
        "morph_path": None,
        "unravelled_cut_leaves_path": "",
        "unravelled_apical_point_path": "",
    }

    validation_function = unravel

    def kwargs(self):
        """ """
        return {"window_half_length": self.window_half_length}

    def inputs(self):
        """ """
        return {
            SmoothDiameters: {"morph_path": "morph_path"},
            CollectAnnotated: {
                "cut_leaves_path": "cut_leaves_path",
                "apical_point_path": "apical_point_path",
            },
        }


class RepairNeurites(ElementValidationTask):
    """RepairNeurites morphologies.

    Using the cut leaves, we recreate missing branches using neuror.main.repair.

    .. todo::

        Currently, axons are not repaired, as they need other axons.

        Improve repair of dendrite by using all dendrites of same type.
    """

    output_columns = {"morph_path": None}
    validation_function = repair
    with_plot = BoolParameter(
        default=False, description=":bool: Save plots with highlighted repaired branches"
    )
    repair_params = luigi.OptionalDictParameter(
        default=None, description=":dict: Repair internal parameters"
    )

    def kwargs(self):
        """ """
        return {"with_plot": self.with_plot, "repair_params": self.repair_params}

    def inputs(self):
        """ """
        return {
            CollectAnnotated: {
                "has_axon": "has_axon",
                "has_basal": "has_basal",
                "has_apical": "has_apical",
            },
            Unravel: {
                "unravelled_apical_point_path": "unravelled_apical_point_path",
                "morph_path": "morph_path",
                "unravelled_cut_leaves_path": "unravelled_cut_leaves_path",
            },
        }


class MakeCollage(SkippableMixin(), SetValidationTask):
    """Make collage plot of morphologies."""

    collage_path = luigi.Parameter(default="collage.pdf", description=":str: Path to collage plot")
    separation = luigi.FloatParameter(default=1500)
    layer_thickness = luigi.ListParameter(default=[700.0, 525.0, 190.0, 353.0, 149.0, 165.0])
    dpi = luigi.IntParameter(default=1000)
    n_morph_per_page = luigi.IntParameter(default=10)

    def kwargs(self):
        """ """
        return {
            "collage_path": self.collage_path,
            "separation": self.separation,
            "layer_thickness": self.layer_thickness,
            "dpi": self.dpi,
            "n_morph_per_page": self.n_morph_per_page,
        }

    validation_function = make_collage

    def inputs(self):
        """ """
        return {RepairNeurites: {"morph_path": "morph_path"}}


class MakeRelease(SetValidationTask):
    """Make a morpology release, with three possible folders: zero-diameter, unravel or repair."""

    zero_diameter_path = luigi.OptionalParameter(
        default=None, description=":str: Path to zero diameter morphologies (not created if None)"
    )
    unravel_path = luigi.OptionalParameter(
        default=None, description=":str: Path to unravel morphologies (not created if None)"
    )
    repair_path = luigi.Parameter(
        default="repaired_release",
        description=":str: Path to repaired morphologies (not created if None)",
    )
    duplicate_layers = luigi.BoolParameter(
        default=True, description=":bool: Duplicate entries with mixed layer mtypes, i.e. L23_PC."
    )

    extensions = [".asc", ".h5", ".swc"]
    output_columns = {}
    for extension in extensions:
        ext = extension[1:]
        output_columns.update(
            {
                f"zero_diameter_morph_db_path_{ext}": None,
                f"unravel_morph_db_path_{ext}": None,
                f"repair_morph_db_path_{ext}": None,
                f"zero_diameter_release_morph_path_{ext}": None,
                f"unravel_release_morph_path_{ext}": None,
                f"repair_release_morph_path_{ext}": None,
                "layer": None,
            }
        )

    validation_function = make_release

    def kwargs(self):
        """ """
        return {
            "zero_diameter_path": self.zero_diameter_path,
            "unravel_path": self.unravel_path,
            "repair_path": self.repair_path,
            "extensions": self.extensions,
            "duplicate_layers": self.duplicate_layers,
        }

    def inputs(self):
        """ """
        return {
            FixZeroDiameters: {"morph_path": "zero_diameter_morph_path"},
            Unravel: {"morph_path": "unravel_morph_path"},
            RepairNeurites: {"morph_path": "repair_morph_path"},
        }


class PlotRepair(SkippableMixin(), ElementValidationTask):
    """Plot the cut leaves on repaired cells."""

    output_columns = {"plot_repair_path": None}
    validation_function = plot_repair

    with_plotly = BoolParameter(default=False, description=":bool: Use Plotly for plotting")

    def kwargs(self):
        """ """
        return {"with_plotly": self.with_plotly}

    def inputs(self):
        """ """
        return {
            Unravel: {"unravelled_cut_leaves_path": "cut_leaves_path"},
            RepairNeurites: {"morph_path": "morph_path"},
        }


class SmoothDiameters(SkippableMixin(True), ElementValidationTask):
    """Smooth diameters with :mod:`diameter_synthesis`.

    We use actual diameters to learn a diameter model used to diametrize the morphology.

    By default, this task is skipped.
    """

    output_columns = {"morph_path": None}

    validation_function = smooth_diameters

    def inputs(self):
        """ """
        return {
            FixZeroDiameters: {"morph_path": "morph_path"},
            CollectAnnotated: {"apical_point_path": "apical_point_path"},
        }


class PlotSmoothDiameters(SkippableMixin(True), ElementValidationTask):
    """Plot smoothed diameters versus originals."""

    output_columns = {"plot_smooth_path": None}
    validation_function = plot_smooth_diameters

    def inputs(self):
        """ """
        return {
            FixZeroDiameters: {"morph_path": "morph_path"},
            SmoothDiameters: {"morph_path": "smooth_morph_path"},
        }



"""Package containing the workflow tasks."""
import os

import luigi_tools.util
from pkg_resources import resource_filename

# Get configuration templates from _templates/config_templates
_template_dir = os.environ.get("MORPHOLOGY_WORKFLOWS_TEMPLATE_DIR")
if _template_dir is None:
    _template_dir = resource_filename(
        "morphology_workflows",
        "_templates/config_templates",
    )
luigi_tools.util.register_templates(directory=_template_dir, name="default.cfg")



"""Annotation functions."""
import json
import logging
import re
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from data_validation_framework.result import ValidationResult
from morph_tool.apical_point import apical_point_position
from morph_tool.transform import rotate
from morphio import SectionType
from morphio.mut import Morphology
from neurom import COLS
from neurom import geom
from neurom import iter_neurites
from neurom import load_morphology
from neurom import view
from neurom.core.types import NeuriteType
from neurom.core.types import tree_type_checker
from neuror.cut_plane.cut_leaves import find_cut_leaves

from morphology_workflows.marker_helper import Marker
from morphology_workflows.marker_helper import MarkerSet

logger = logging.getLogger(__name__)
matplotlib.use("Agg")


def _get_bbox(neuron):
    """Get bbox for all dendrite except axon."""
    min_x, max_x, min_y, max_y = 1e10, -1e10, 1e10, -1e10
    for neurite in neuron.neurites:
        max_x = max(max_x, np.max(neurite.points[:, 0]))
        min_x = min(min_x, np.min(neurite.points[:, 0]))
        max_y = max(max_y, np.max(neurite.points[:, 1]))
        min_y = min(min_y, np.min(neurite.points[:, 1]))
    return min_x, max_x, min_y, max_y


def _get_bbox_no_axon(neuron):
    """Get bbox for all dendrite except axon."""
    min_x, max_x, min_y, max_y = 1e10, -1e10, 1e10, -1e10
    for neurite in neuron.neurites:
        if neurite.type != SectionType.axon:
            max_x = max(max_x, np.max(neurite.points[:, 0]))
            min_x = min(min_x, np.min(neurite.points[:, 0]))
            max_y = max(max_y, np.max(neurite.points[:, 1]))
            min_y = min(min_y, np.min(neurite.points[:, 1]))
    return min_x, max_x, min_y, max_y


def assign_mtypes(row, data_dir, mtype_regex=None):  # pylint: disable=unused-argument
    """Assign mtype to morphologies."""
    if not mtype_regex:
        return ValidationResult(is_valid=True, mtype=row.mtype, comment="No regex provided!")
    if not hasattr(row, "mtype") or not isinstance(row.mtype, str):
        return ValidationResult(is_valid=True, mtype=row.mtype, comment="No  mtype data!")

    mtype_re = [re.compile(_re) for _re in mtype_regex]
    _matches = [_re.findall(row.mtype)[0] for _re in mtype_re if _re.match(row.mtype)]
    if len(_matches) > 1:
        logger.warning("Found several mtype names: %s, so we take the first", _matches)
    if len(_matches) == 0:
        return ValidationResult(is_valid=False, comment=f"provided mtype is not valid: {row.mtype}")
    return ValidationResult(is_valid=True, mtype=_matches[0])


def calculate_y_extent(morph, neurite_type):
    """Find min/max y value of morphology based on neurite_type.

    Taken from morphology_repair_workflow.placement_annotations.py
    """
    total_min = float("inf")
    total_max = float("-inf")
    for n in iter_neurites(morph, filt=neurite_type):
        min_, max_ = geom.bounding_box(n)
        total_min = min(min_[COLS.Y], total_min)
        total_max = max(max_[COLS.Y], total_max)
    return total_min, total_max


SEGMENT_TO_NEURITE = {
    "axon": tree_type_checker(NeuriteType.axon),
    "dendrite": tree_type_checker(NeuriteType.basal_dendrite, NeuriteType.apical_dendrite),
}


def compute_hard_limits(
    row,
    data_dir,
    skip_axon=False,
    dendrite_hard_limit="L1_hard_limit",
    axon_hard_limit="L1_axon_hard_limit",
):
    """Compute hard limits."""
    morph = load_morphology(row.morph_path)

    rules = {}
    y_min, y_max = calculate_y_extent(morph, SEGMENT_TO_NEURITE["dendrite"])
    rules[dendrite_hard_limit] = dict(y_min=y_min, y_max=y_max)
    if not skip_axon:
        y_min, y_max = calculate_y_extent(morph, SEGMENT_TO_NEURITE["axon"])
        rules[axon_hard_limit] = dict(y_min=y_min, y_max=y_max)

    _markers = []
    for name, rule in rules.items():
        for tpe in ["y_min", "y_max"]:
            _markers.append(
                Marker(
                    name + "_" + tpe,
                    "axis",
                    [[0.0, rule[tpe], 0.0], [1.0, rule[tpe], 0.0]],
                    morph_name=row.name,
                    morph_path=row.morph_path,
                )
            )

    hard_limit_path = Path(data_dir, row.name + ".yaml")
    MarkerSet.from_markers(_markers).save(filename=hard_limit_path)
    return ValidationResult(is_valid=True, hard_limit_path=hard_limit_path)


def _labeled_line(ax, name, y_pos, color, pos="center"):
    """Draw a line to the axis, returning the line and the annotation."""
    x_pos = {"center": 0, "left": ax.dataLim.bounds[0]}[pos]
    line = ax.axhline(y_pos, c=color)
    anno = ax.annotate(name, xy=(0, y_pos), xytext=(x_pos, y_pos + 3))
    return line, anno


def plot_hard_limits(row, data_dir, with_plotly=True):
    """Plotting placement annotations on morphologies.

    TODO: update the plotting without plotly.
    """
    plot_path = None
    if not row.isnull()["hard_limit_path"]:
        if with_plotly:
            plot_path = (data_dir / row.name).with_suffix(".html")
            MarkerSet.from_file(row.hard_limit_path).plot(filename=plot_path)
        # else:
        #    pa = PlacementAnnotation.load(Path(row.hard_limit_path).with_suffix(""))
        #    neuron = load_morphology(row.morph_path)
        #    plt.figure()
        #    ax = plt.gca()
        #    for rule, data in pa.items():
        #        _labeled_line(ax, rule + " min", data["y_min"], color="g")
        #        _labeled_line(ax, rule + " max", data["y_max"], color="g")
        #
        #    view.plot_morph(neuron, ax, realistic_diameters=True)
        #
        #    plt.axis(_get_bbox(neuron))
        #    plot_path = (data_dir / row.name).with_suffix(".pdf")
        #    plt.savefig(plot_path)
        #    plt.close()

    return ValidationResult(is_valid=True, plot_hard_limit_path=plot_path)


def find_apical_point(row, data_dir, tuft_percent=20):  # pylint: disable=unused-argument
    """Find apical point."""
    if row.has_apical:
        neuron = Morphology(row.morph_path)
        apical_point = apical_point_position(neuron, tuft_percent=tuft_percent)
        if apical_point is not None:
            marker = Marker(
                "apical point",
                "points",
                apical_point,
                morph_name=row.name,
                morph_path=row.morph_path,
                plot_style={"color": "purple"},
            )
            apical_point_path = (data_dir / Path(row.morph_path).name).with_suffix(".yaml")
            MarkerSet.from_markers([marker]).save(filename=apical_point_path)
            return ValidationResult(is_valid=True, apical_point_path=apical_point_path)
    return ValidationResult(is_valid=True, apical_point_path=None)


def plot_apical_point(row, data_dir, with_plotly=True):
    """Plotting apical points on morphologies."""
    plot_path = None
    if row.has_apical and not row.isnull()["apical_point_path"]:
        if with_plotly:
            plot_path = str(data_dir / row.name) + ".html"
            MarkerSet.from_file(row.apical_point_path).plot(filename=plot_path)
        else:  # pragma: no cover
            plot_path = str(data_dir / row.name) + ".pdf"
            neuron = load_morphology(row.morph_path)
            plt.figure()
            apical_point = MarkerSet.from_file(row.apical_point_path)
            view.plot_morph(neuron, plt.gca(), realistic_diameters=True, soma_outline=False)
            plt.scatter(*apical_point.markers[0].data[:2], c="g", s=20)
            plt.axis(_get_bbox_no_axon(neuron))
            plt.axis("equal")
            plt.savefig(plot_path)
            plt.close()
    return ValidationResult(is_valid=True, plot_apical_point_path=plot_path)


def detect_cut_leaves(row, data_dir, bin_width=15, percentile_threshold=75):
    """Detect cut leaves and save status in df."""
    morph = Morphology(row.morph_path)
    rotation_matrix = np.eye(3)
    if not row.isnull()["rotation_matrix"]:
        rotation_matrix = np.array(json.loads(row.rotation_matrix)).T
        rotate(morph, rotation_matrix)
    morph = morph.as_immutable()

    if not morph.root_sections:
        raise ValueError("Can not search for cut leaves for a neuron with no neurites.")

    leaves, qualities = find_cut_leaves(
        morph,
        bin_width=bin_width,
        searched_axes=("Z",),
        searched_half_spaces=(-1, 1),
        percentile_threshold=percentile_threshold,
    )

    if len(leaves) > 0:
        leaves = leaves.dot(rotation_matrix)
        marker = Marker(
            "cut leaves",
            "points",
            leaves,
            morph_name=row.name,
            morph_path=row.morph_path,
            plot_style={"color": "green"},
        )
        plane_path = (data_dir / Path(row.morph_path).name).with_suffix(".yaml")
        MarkerSet.from_markers([marker]).save(filename=plane_path)
        return ValidationResult(
            is_valid=True,
            cut_leaves_path=plane_path,
            cut_qualities=json.dumps(qualities),
        )
    return ValidationResult(
        is_valid=True,
        cut_leaves_path=None,
        cut_qualities=None,
        comment="No cut leaves found",
    )


def plot_cut_leaves(row, data_dir, with_plotly=True):
    """Plotting cut leaves on morphologies."""
    markers = None
    if not row.isnull()["cut_leaves_path"]:
        markers = MarkerSet.from_file(row.cut_leaves_path)

    plot_path = None
    if with_plotly:
        if markers:
            plot_path = (data_dir / row.name).with_suffix(".html")
            markers.plot(filename=plot_path)
    else:  # pragma: no cover
        plt.figure()
        plot_path = (data_dir / row.name).with_suffix(".pdf")
        neuron = load_morphology(row.morph_path)
        view.plot_morph(
            neuron, plt.gca(), realistic_diameters=False, plane="xz", soma_outline=False
        )

        if markers:
            points = markers.markers[0].data.T[[0, 2]]
            plt.scatter(*points, color="g", s=5)

        plt.gca().axis("equal")
        plt.title(f"{row.name}")
        plt.suptitle(f"probability = {row.cut_qualities}", fontsize=8)
        plt.savefig(plot_path)
        plt.close()

    return ValidationResult(is_valid=True, plot_cut_leaves_path=plot_path)



"""Test curation functions."""
import json

import numpy as np
import pandas as pd
import pytest
from morphio.mut import Morphology
from numpy.testing import assert_array_almost_equal

from morphology_workflows import curation

from . import create_morphology
from . import create_morphology_file


class Test_fix_soma_radius:
    """Test the function curation.fix_soma_radius()."""

    def test_nothing_to_fix(self):
        """No point in soma."""
        morph = create_morphology(
            """
            1 1 0 0 0 1. -1
            2 2 1 0 0 1. 1
            3 2 2 0 0 1. 2
            4 2 3 0 0 1. 3
            """,
            "swc",
        )

        former_radius, new_radius = curation.fix_soma_radius(morph)
        assert former_radius == 1
        assert new_radius is None

    def test_first_point_in_soma(self):
        """First point in soma."""
        morph = create_morphology(
            """
            1 1 0 0 0 1.5 -1
            2 2 1 0 0 1. 1
            3 2 2 0 0 1. 2
            4 2 3 0 0 1. 3
            """,
            "swc",
        )

        former_radius, new_radius = curation.fix_soma_radius(morph)
        assert former_radius == 1.5
        assert new_radius is None

    def test_last_point_in_soma(self):
        """All points in soma."""
        morph = create_morphology(
            """
            1 1 0 0 0 10 -1
            2 2 1 0 0 1. 1
            3 2 2 0 0 1. 2
            4 2 3 0 0 1. 3
            """,
            "swc",
        )

        former_radius, new_radius = curation.fix_soma_radius(morph)
        assert former_radius == 10
        assert new_radius == 2.5

    def test_last_point_closer_than_before_last(self):
        """All points in soma and the last points is closer than the before last point."""
        morph = create_morphology(
            """
            1 1 0 0 0 10 -1
            2 2 1 0 0 1. 1
            3 2 3 0 0 1. 2
            4 2 2 0 0 1. 3
            """,
            "swc",
        )

        former_radius, new_radius = curation.fix_soma_radius(morph)
        assert former_radius == 10
        assert new_radius == 2.5

    def test_two_points(self):
        """Only 2 points and both in soma."""
        morph = create_morphology(
            """
            1 1 0 0 0 10 -1
            2 2 2 0 0 1. 1
            3 2 3 0 0 1. 2
            """,
            "swc",
        )

        former_radius, new_radius = curation.fix_soma_radius(morph)
        assert former_radius == 10
        assert new_radius == 2.5

    def test_two_points_with_last_closer(self):
        """Only 2 points and the last points is closer than the before last point."""
        morph = create_morphology(
            """
            1 1 0 0 0 10 -1
            2 2 3 0 0 1. 1
            3 2 2 0 0 1. 2
            """,
            "swc",
        )

        former_radius, new_radius = curation.fix_soma_radius(morph)
        assert former_radius == 10
        assert new_radius == 2.5


class TestFixRootSections:
    """Test the function curation.fix_root_sections()."""

    def test_nothing_to_fix(self):
        """The first section is long enough."""
        morph = create_morphology(
            """
            1 1 0 0 0 1. -1
            2 2 1 0 0 1. 1
            3 2 1.5 0 0 1. 2
            4 2 2 0 0 1. 3
            5 2 3 0 0 1. 3
            """,
            "swc",
        )

        curation.fix_root_section(morph)
        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.5, 0.0, 0.0, 1.0],
                [1.5, 0.0, 0.0, 1.0],
                [2.0, 0.0, 0.0, 1.0],
                [1.5, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
            ],
        )

    def test_small_section(self):
        """The first section is too small."""
        morph = create_morphology(
            """
            1 1 0 0 0 1. -1
            2 2 1 0 0 1. 1
            3 2 1.05 0 0 1. 2
            4 2 2 0 0 1. 3
            5 2 3 0 0 1. 3
            """,
            "swc",
        )

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [2.0, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
            ],
        )

        curation.fix_root_section(morph, min_length=0.5)

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.5, 0.0, 0.0, 1.0],
                [1.5, 0.0, 0.0, 1.0],
                [2.0, 0.0, 0.0, 1.0],
                [1.5, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
            ],
        )

    def test_small_section_and_small_child(self):
        """The first section is too small and one of its children is also too small."""
        morph = create_morphology(
            """
            1 1 0 0 0 1. -1
            2 2 1 0 0 1. 1
            3 2 1.05 0 0 1. 2
            4 2 1.06 0 0 1. 3
            5 2 3 0 0 1. 4
            6 2 3 1 0 1. 3
            """,
            "swc",
        )

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [1.06, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [3.0, 1.0, 0.0, 1.0],
            ],
        )

        curation.fix_root_section(morph, min_length=0.5)

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.486032, 0.117358, 0.0, 1.0],
                [1.486032, 0.117358, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
                [1.486032, 0.117358, 0.0, 1.0],
                [3.0, 1.0, 0.0, 1.0],
            ],
        )

    def test_two_small_sections(self):
        """The two first sections are too small."""
        morph = create_morphology(
            """
            1 1 0 0 0 1. -1
            2 2 1 0 0 1. 1
            3 2 1.05 0 0 1. 2
            4 2 1.1 0 0 1. 3
            5 2 3 0.5 0 1. 4
            6 2 3 0 0 1. 4
            7 2 3 1 0 1. 3
            """,
            "swc",
        )

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [1.1, 0.0, 0.0, 1.0],
                [1.1, 0.0, 0.0, 1.0],
                [3.0, 0.5, 0.0, 1.0],
                [1.1, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [3.0, 1.0, 0.0, 1.0],
            ],
        )

        curation.fix_root_section(morph, min_length=0.5)

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.4777161, 0.1476055, 0.0, 1.0],
                [1.4777161, 0.1476055, 0.0, 1.0],
                [1.5277162, 0.1476055, 0.0, 1.0],
                [1.5277162, 0.1476055, 0.0, 1.0],
                [3.0, 0.5, 0.0, 1.0],
                [1.5277162, 0.1476055, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
                [1.4777161, 0.1476055, 0.0, 1.0],
                [3.0, 1.0, 0.0, 1.0],
            ],
        )

    def test_three_small_sections(self):
        """The three first sections are too small."""
        morph = create_morphology(
            """
            1 1 0 0 0 1. -1
            2 2 1 0 0 1. 1
            3 2 1.05 0 0 1. 2
            4 2 1.1 0 0 1. 3
            5 2 1.15 0 0 1. 4
            6 2 3 0.5 0 1. 5
            7 2 3 0 0 1. 5
            8 2 3 0.75 0 1. 4
            9 2 3 1 0 1. 3
            """,
            "swc",
        )

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [1.1, 0.0, 0.0, 1.0],
                [1.1, 0.0, 0.0, 1.0],
                [1.15, 0.0, 0.0, 1.0],
                [1.15, 0.0, 0.0, 1.0],
                [3.0, 0.5, 0.0, 1.0],
                [1.15, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
                [1.1, 0.0, 0.0, 1.0],
                [3.0, 0.75, 0.0, 1.0],
                [1.05, 0.0, 0.0, 1.0],
                [3.0, 1.0, 0.0, 1.0],
            ],
        )

        curation.fix_root_section(morph, min_length=0.5)

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.4679317, 0.17618132, 0.0, 1.0],
                [1.4679317, 0.17618132, 0.0, 1.0],
                [1.5179318, 0.17618132, 0.0, 1.0],
                [1.5179318, 0.17618132, 0.0, 1.0],
                [1.5679318, 0.17618132, 0.0, 1.0],
                [1.5679318, 0.17618132, 0.0, 1.0],
                [3.0, 0.5, 0.0, 1.0],
                [1.5679318, 0.17618132, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
                [1.5179318, 0.17618132, 0.0, 1.0],
                [3.0, 0.75, 0.0, 1.0],
                [1.4679317, 0.17618132, 0.0, 1.0],
                [3.0, 1.0, 0.0, 1.0],
            ],
        )

    def test_three_zero_sections(self):
        """The three first sections have 0 lengths."""
        morph = create_morphology(
            """
            1 1 0 0 0 1. -1
            2 2 1. 0 0 1. 1
            3 2 1. 0 0 1. 2
            4 2 1. 0 0 1. 3
            5 2 1. 0 0 1. 4
            6 2 3 0.5 0 1. 5
            7 2 3 0 0 1. 5
            8 2 3 0.75 0 1. 4
            9 2 3 1 0 1. 3
            """,
            "swc",
        )

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [3.0, 0.5, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [3.0, 0.75, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [3.0, 1.0, 0.0, 1.0],
            ],
        )

        curation.fix_root_section(morph)

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.0000938, 3.4490615e-05, 0.0, 1.0],
                [1.0000938, 3.4490615e-05, 0.0, 1.0],
                [1.0000938, 3.4490615e-05, 0.0, 1.0],
                [1.0000938, 3.4490615e-05, 0.0, 1.0],
                [3.0, 5.0e-01, 0.0, 1.0],
                [1.0000938, 3.4490615e-05, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
                [1.0000938, 3.4490615e-05, 0.0, 1.0],
                [3.0, 7.5e-01, 0.0, 1.0],
                [1.0000938, 3.4490615e-05, 0.0, 1.0],
                [3.0, 1.0, 0.0, 1.0],
            ],
        )

    def test_zero_sections_no_child(self):
        """The first sections has 0 length and no child."""
        morph = create_morphology(
            """
            1 1 0 0 0 1. -1
            2 2 1. 0 0 1. 1
            3 2 1. 0 0 1. 2
            """,
            "swc",
        )

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
            ],
        )

        curation.fix_root_section(morph)

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.0001, 0.0, 0.0, 1.0],
            ],
        )

    def test_zero_sections_no_child_and_overlap_soma(self):
        """The first sections has 0 length and no child."""
        morph = create_morphology(
            """
            1 1 1. 0 0 1. -1
            2 2 1. 0 0 1. 1
            3 2 1. 0 0 1. 2
            4 2 2. 0 0 1. 1
            5 2 3. 0 0 1. 4
            """,
            "swc",
        )

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [2.0, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
            ],
        )

        curation.fix_root_section(morph)

        assert_array_almost_equal(
            morph.points,
            [
                [2.0, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
            ],
        )

    def test_too_small_min_length(self, caplog):
        """The min length is too small compared to the precision to move a point."""
        morph = create_morphology(
            """
            1 1 0 0 0 1. -1
            2 2 1 0 0 1. 1
            3 2 1 0 0 1. 2
            4 2 2 0 0 1. 3
            5 2 3 0 0 1. 3
            """,
            "swc",
        )

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [2.0, 0.0, 0.0, 1.0],
                [1.0, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
            ],
        )

        caplog.clear()
        with caplog.at_level("DEBUG"):
            curation.fix_root_section(morph, min_length=1e-9)

        assert_array_almost_equal(
            morph.points,
            [
                [1.0, 0.0, 0.0, 1.0],
                [1.0000001, 0.0, 0.0, 1.0],
                [1.0000001, 0.0, 0.0, 1.0],
                [2.0, 0.0, 0.0, 1.0],
                [1.0000001, 0.0, 0.0, 1.0],
                [3.0, 0.0, 0.0, 1.0],
            ],
        )

        assert caplog.messages == [
            (
                "The min_length was too small to move the point [1.0, 0.0, 0.0] so "
                "it was moved to [1.0000001, 0.0, 0.0]"
            )
        ]


class TestCheckNeurites:
    """Test the function curation.check_neurites()."""

    @pytest.fixture
    def res_path(self, tmpdir):
        """Result to which the result morphologies are exported."""
        path = tmpdir / "res"
        path.mkdir()
        return path

    @pytest.fixture
    def simple_morph(self, tmpdir):
        """A simple morphology used for testing."""
        return create_morphology_file(
            """
            1 1 0 0 0 1 -1
            5 3 -1 0 0 1. 1
            6 3 -2 0 0 1. 5
            7 3 -3 0 0 1. 6
            8 4 0 1 0 1. 1
            9 4 0 6 0 1. 8
            10 4 1 7 0 1. 9
            """,
            "swc",
            tmpdir / "morph.swc",
        )

    def test_default(self, simple_morph, res_path):
        """Check neurites with default options."""
        row = pd.Series({"morph_path": simple_morph}, name="test_name")
        res = curation.check_neurites(
            row,
            res_path,
        )

        res_morph = Morphology(res["morph_path"])
        expected = {
            0: [[-1, 0, 0], [-2, 0, 0], [-3, 0, 0]],
            1: [[0, 1, 0], [0, 6, 0], [1, 7, 0]],
            2: [[0, 0, 0], [0, -100, 0]],
        }
        for i, j in zip(res_morph.sections.items(), expected.items()):
            assert i[0] == j[0]
            assert_array_almost_equal(i[1].points, j[1])

    def test_no_soma(self, simple_morph, res_path):
        """Check neurites with default options on a morph without any soma."""
        morph = Morphology(simple_morph)
        morph.soma.points = np.array([], dtype=morph.soma.points.dtype).reshape((0, 3))
        morph.soma.diameters = np.array([], dtype=morph.soma.diameters.dtype)
        morph.write(simple_morph)
        row = pd.Series({"morph_path": simple_morph}, name="test_name")

        # Test with spherical soma
        res = curation.check_neurites(
            row,
            res_path,
        )

        res_morph = Morphology(res["morph_path"])
        expected = {
            0: [[-1, 0, 0], [-2, 0, 0], [-3, 0, 0]],
            1: [[0, 1, 0], [0, 6, 0], [1, 7, 0]],
            2: [[0, 0, 0], [0, -100, 0]],
        }
        for i, j in zip(res_morph.sections.items(), expected.items()):
            assert i[0] == j[0]
            assert_array_almost_equal(i[1].points, j[1])
        assert_array_almost_equal(res_morph.soma.points, [[-0.5, 0.5, 0]])

        # Test with contour soma
        res = curation.check_neurites(
            row,
            res_path,
            mock_soma_type="contour",
        )

        res_morph = Morphology(res["morph_path"])
        expected = {
            0: [[-1, 0, 0], [-2, 0, 0], [-3, 0, 0]],
            1: [[0, 1, 0], [0, 6, 0], [1, 7, 0]],
            2: [[0, 0, 0], [0, -100, 0]],
        }
        for i, j in zip(res_morph.sections.items(), expected.items()):
            assert i[0] == j[0]
            assert_array_almost_equal(i[1].points, j[1])
        assert_array_almost_equal(
            res_morph.soma.points,
            [
                [-1, 0, 0],
                [0, 1, 0],
            ],
        )

    def test_no_mock_but_stub(self, simple_morph, res_path):
        """Check neurites with no mock soma but with stub axon."""
        row = pd.Series({"morph_path": simple_morph}, name="test_name")
        res = curation.check_neurites(
            row,
            res_path,
            mock_soma_type=None,
            ensure_stub_axon=True,
        )

        res_morph = Morphology(res["morph_path"])
        expected = {
            0: [[-1, 0, 0], [-2, 0, 0], [-3, 0, 0]],
            1: [[0, 1, 0], [0, 6, 0], [1, 7, 0]],
            2: [[0, 0, 0], [0, -100, 0]],
        }
        for i, j in zip(res_morph.sections.items(), expected.items()):
            assert i[0] == j[0]
            assert_array_almost_equal(i[1].points, j[1])


class TestFixNeuritesInSoma:
    """Test the function curation.fix_neurites_in_soma()."""

    @pytest.fixture
    def res_path(self, tmpdir):
        """Result to which the result morphologies are exported."""
        path = tmpdir / "res"
        path.mkdir()
        return path

    @pytest.fixture
    def simple_morph(self, tmpdir):
        """A simple morphology used for testing."""
        return create_morphology_file(
            """
            1 1 0 0 0 10 -1
            2 2 1 0 0 1. 1
            3 2 2 0 0 1. 2
            4 2 6 0 0 1. 3
            5 3 -1 0 0 1. 1
            6 3 -2 0 0 1. 5
            7 3 -3 0 0 1. 6
            8 4 0 1 0 1. 1
            9 4 0 6 0 1. 8
            10 4 1 7 0 1. 9
            """,
            "swc",
            tmpdir / "morph.swc",
        )

    def test_default(self, simple_morph, res_path):
        """Align with default options."""
        row = pd.Series({"morph_path": simple_morph}, name="test_name")
        res = curation.fix_neurites_in_soma(
            row,
            res_path,
        )

        res_morph = Morphology(res["morph_path"])
        expected = {
            0: [
                [2.5, 0, 0],
                [6, 0, 0],
            ],
            1: [
                [-2.5, 0, 0],
                [-3, 0, 0],
            ],
            2: [
                [0, 2.5, 0],
                [0, 6, 0],
                [1, 7, 0],
            ],
        }
        for i, j in zip(res_morph.sections.items(), expected.items()):
            assert i[0] == j[0]
            assert_array_almost_equal(i[1].points, j[1])


class TestAlign:
    """Test the function curation.align()."""

    @pytest.fixture
    def res_path(self, tmpdir):
        """Result to which the result morphologies are exported."""
        path = tmpdir / "res"
        path.mkdir()
        return path

    @pytest.fixture
    def simple_morph(self, tmpdir):
        """A simple morphology used for testing."""
        return create_morphology_file(
            """
            1 1 0 0 0 1. -1
            2 2 1 0 0 1. 1
            3 2 2 0 0 1. 2
            4 2 3 0 0 1. 3
            5 3 0 1 0 1. 1
            6 3 0 2 0 1. 5
            7 3 0 3 0 1. 6
            8 4 -1 0 0 1. 1
            9 4 -2 0 0 1. 8
            10 4 -3 0 0 1. 9
            """,
            "swc",
            tmpdir / "morph.swc",
        )

    def check_res(self, res, expected):
        """Compare the result points to the expected ones."""
        res_morph = Morphology(res["morph_path"])
        for i, j in zip(res_morph.sections.items(), expected.items()):
            assert i[0] == j[0]
            assert_array_almost_equal(i[1].points, j[1])

    def test_default(self, simple_morph, res_path):
        """Align with default options."""
        row = pd.Series({"morph_path": simple_morph}, name="test_name")
        res = curation.align(
            row,
            res_path,
        )

        expected = {
            0: [[0, -1, 0], [0, -2, 0], [0, -3, 0]],
            1: [[1, 0, 0], [2, 0, 0], [3, 0, 0]],
            2: [[0, 1, 0], [0, 2, 0], [0, 3, 0]],
        }
        self.check_res(res, expected)

    def test_neurite_type(self, simple_morph, res_path):
        """Align with a given neurite type."""
        row = pd.Series({"morph_path": simple_morph}, name="test_name")
        res = curation.align(
            row,
            res_path,
            neurite_type="axon",
        )

        expected = {
            0: [[0, 1, 0], [0, 2, 0], [0, 3, 0]],
            1: [[-1, 0, 0], [-2, 0, 0], [-3, 0, 0]],
            2: [[0, -1, 0], [0, -2, 0], [0, -3, 0]],
        }
        self.check_res(res, expected)

    def test_given_direction(self, simple_morph, res_path):
        """Align with a given direction."""
        row = pd.Series({"morph_path": simple_morph}, name="test_name")
        res = curation.align(
            row,
            res_path,
            direction=[1, 0, 0],
        )

        expected = {
            0: [[1, 0, 0], [2, 0, 0], [3, 0, 0]],
            1: [[0, 1, 0], [0, 2, 0], [0, 3, 0]],
            2: [[-1, 0, 0], [-2, 0, 0], [-3, 0, 0]],
        }
        self.check_res(res, expected)

    def test_custom_method(self, tmpdir, simple_morph, res_path):
        """Align with a custom rotation method."""
        custom_orientation_json_path = tmpdir / "custom_orientation.json"
        with open(custom_orientation_json_path, mode="w", encoding="utf-8") as f:
            json.dump(
                {
                    "test_name": [-np.sqrt(2) / 2, np.sqrt(2) / 2, 0],
                },
                f,
            )
        row = pd.Series({"morph_path": simple_morph}, name="test_name")
        res = curation.align(
            row,
            res_path,
            method="custom",
            custom_orientation_json_path=custom_orientation_json_path,
        )

        expected = {
            0: [
                [np.sqrt(2) / 2, -np.sqrt(2) / 2, 0],
                [np.sqrt(2), -np.sqrt(2), 0],
                [3 * np.sqrt(2) / 2, -3 * np.sqrt(2) / 2, 0],
            ],
            1: [
                [np.sqrt(2) / 2, np.sqrt(2) / 2, 0],
                [np.sqrt(2), np.sqrt(2), 0],
                [3 * np.sqrt(2) / 2, 3 * np.sqrt(2) / 2, 0],
            ],
            2: [
                [-np.sqrt(2) / 2, np.sqrt(2) / 2, 0],
                [-np.sqrt(2), np.sqrt(2), 0],
                [-3 * np.sqrt(2) / 2, 3 * np.sqrt(2) / 2, 0],
            ],
        }
        self.check_res(res, expected)

        # Test with unknown morphology name => should not be rotated
        row = pd.Series({"morph_path": simple_morph}, name="unknown_name")
        res = curation.align(
            row,
            res_path,
            method="custom",
            custom_orientation_json_path=custom_orientation_json_path,
        )

        expected = {i: j.points.tolist() for i, j in Morphology(simple_morph).sections.items()}
        self.check_res(res, expected)

        # Test with custom method but no custom_orientation_json_path
        with pytest.raises(
            ValueError,
            match="Provide a custom_orientation_json_path parameter when method=='custom'",
        ):
            res = curation.align(
                row,
                res_path,
                method="custom",
            )



"""Module to read/write/plot morphology markers."""
import logging

import numpy as np
import yaml
from neurom import load_morphology
from neurom.geom import bounding_box
from plotly_helper.neuron_viewer import NeuronBuilder
from plotly_helper.object_creator import scatter
from plotly_helper.object_creator import scatter_line
from plotly_helper.object_creator import vector

logger = logging.getLogger(__name__)


class MarkerSet:
    """Container of several markers for a single morphology."""

    def __init__(self):
        """ """
        self.markers = None
        self.morph_name = None
        self.morph_path = None

    @classmethod
    def from_markers(cls, markers):
        """Load from a list of Marker object."""
        obj = cls()

        if len({marker.morph_name for marker in markers}) > 1:
            raise Exception("Markers of different morphologies were provided.")
        obj.morph_name = markers[0].morph_name

        if len({marker.morph_path for marker in markers}) > 1:
            raise Exception("Markers of different morphology paths were provided.")
        obj.morph_path = markers[0].morph_path

        obj.markers = cls.check_labels(markers)
        return obj

    @staticmethod
    def check_labels(markers):
        """Ensure label names are different."""
        labels = {}
        for marker in markers:
            if marker.label in labels:
                labels[marker.label] += 1
                marker.label += "_" + str(labels[marker.label])
            else:
                labels[marker.label] = 0
        return markers

    @classmethod
    def from_file(cls, marker_path):
        """Load maker yaml file."""
        with open(marker_path, "r", encoding="utf-8") as f:
            return MarkerSet.from_dicts(yaml.safe_load(f))

    @classmethod
    def from_dicts(cls, markers):
        """Load marker from dict, where markers["markers"] is a list of dict markers."""
        obj = cls()
        obj.markers = cls.check_labels([Marker(**marker) for marker in markers["markers"]])
        obj.morph_name = markers.get("morph_name", None)
        obj.morph_path = markers.get("morph_path", None)
        return obj

    @staticmethod
    def _plot_points_marker(builder, marker):
        """Plot point markers."""
        data = np.array(marker.data)
        if len(np.shape(data)) == 1:
            data = data[np.newaxis]
        plot_data = scatter(data[:3], name=f"{marker.label}", **marker.plot_style)
        builder.helper.add_data({f"{marker.label}": plot_data})

    @staticmethod
    def _plot_line_marker(builder, marker):
        """Plot line markers."""
        plot_data = scatter_line(marker.data, name=f"{marker.label}", **marker.plot_style)
        builder.helper.add_data({f"{marker.label}": plot_data})

    @staticmethod
    def _plot_axis_marker(builder, marker, margin=10):
        """Plot axis marker.

        This marker is a line crossing the whole bbox.
        """
        point_x = np.array(marker.data[0])
        point_y = np.array(marker.data[1])

        def _bbox(neuron, margin=10):
            """Get enlarged bbox of neuron for plotting."""
            bbox = bounding_box(neuron)
            bbox[0] -= margin * np.ones(3)
            bbox[1] += margin * np.ones(3)
            return bbox

        if not builder.neuron.neurites:
            raise ValueError("Can not plot axis marker for a neuron with no neurites.")
        bbox = _bbox(builder.neuron, margin=margin)

        def _get_lim(direction=1):
            """Find point near bbox limit in directiion of (point_y - point_x)."""
            fac = 1.0
            point = point_x.copy()
            while (point > bbox[0]).all() and (point < bbox[1]).all():
                fac += 10.0
                point = point_x - direction * fac * (point_y - point_x)
            return point

        plot_data = vector(_get_lim(1), _get_lim(-1), name=f"{marker.label}", **marker.plot_style)
        builder.helper.add_data({f"{marker.label}": plot_data})

    def plot(self, filename="markers.html", with_plotly=True, plane="3d", **kwargs):
        """Plot morphology with markers."""
        if with_plotly:

            neuron = load_morphology(self.morph_path)
            builder = NeuronBuilder(neuron, plane, line_width=4, title=f"{self.morph_name}")

            if "auto_open" not in kwargs:
                kwargs["auto_open"] = False

            for marker in self.markers:
                if marker.marker_type == "points":
                    self._plot_points_marker(builder, marker)
                elif marker.marker_type == "axis":
                    self._plot_axis_marker(builder, marker)
                elif marker.marker_type == "line":
                    self._plot_line_marker(builder, marker)
                else:
                    logger.info("marker type %s not understood", marker.type)
            builder.plot(filename=str(filename), **kwargs)
        else:
            raise NotImplementedError("Only plotly available for now")

    def to_dicts(self):
        """Create a list of dicts from marker class."""
        out_dict = {
            "morph_name": self.morph_name,
            "morph_path": self.morph_path,
            "markers": [marker.to_dict()["marker"] for marker in self.markers],
        }
        return out_dict

    def save(self, filename=None, mode="a"):
        """Save marker data.

        Args:
            filename (str): if provided,  path to marker file
            mode (str): writing mode, 'a' or 'w'

        TODO: if data is large, write in an h5/pickle for speedup
        """
        if filename is None:
            filename = self.morph_name.with_suffix(".yaml")
        with open(filename, mode, encoding="utf-8") as f:
            yaml.dump(self.to_dicts(), f)


class Marker:
    """Container of a single marker."""

    def __init__(
        self,
        label,
        marker_type,
        data,
        morph_name=None,
        morph_path=None,
        plot_style=None,
        extra_data=None,
    ):
        """Create a marker.

        Args:
            label (str): label of marker
            marker_type (str): marker types (points, line, plane, etc...)
            data (str): markere data
            morph_name (str): name of corresponding morphology
            morph_path (str): path to corresponding morphology
            plot_style (dict): custom dict for plotting arguments
            extra_data (any): additional marker information to keep (better to be yaml-friendly)
        """
        self.label = label
        self.marker_type = marker_type
        self.data = np.array(data, dtype=float)
        self.morph_name = morph_name
        self.morph_path = morph_path
        self._set_plot_style(plot_style)
        self.extra_data = extra_data

        self._check_valid()

    def _set_plot_style(self, plot_style):
        """Set plotting style."""
        self.plot_style = {"showlegend": True}
        if self.marker_type == "points":
            self.plot_style = {"color": "black", "width": 20}
        if self.marker_type == "line":
            self.plot_style = {"color": "black", "width": 1, "marker_size": 0.0001}
        if plot_style is not None:
            self.plot_style.update(plot_style)

    def _check_valid(self):
        """Check if the given marker data is valid."""
        valid = False
        if self.marker_type == "points":
            valid = len(np.shape(self.data)) == 2 or len(np.shape(self.data)) == 1
        if self.marker_type == "axis":
            valid = len(np.shape(self.data)) == 2
        if self.marker_type == "line":
            valid = len(np.shape(self.data)) == 2
        if not valid:
            raise Exception(f"Marker {self.marker_type} is not valid with data {self.data}.")

    @property
    def list_data(self):
        """Returns data without numpy types."""
        return self.data.tolist()

    def to_dict(self):
        """Create a dict from marker class."""
        return {
            "morph_name": self.morph_name,
            "morph_path": self.morph_path,
            "marker": {
                "label": self.label,
                "marker_type": self.marker_type,
                "data": self.list_data,
                "plot_style": self.plot_style,
                "extra_data": self.extra_data,
            },
        }



"""Placeholders functions."""
import logging
import warnings
from pathlib import Path
from typing import Optional

import neurom
import numpy as np
import pandas as pd
from neurom.apps import morph_stats
from pkg_resources import resource_filename

logger = logging.getLogger(__name__)


DEFAULT_CONFIG = {
    "neurite": {
        "aspect_ratio": {
            "kwargs": [
                {"projection_plane": "xy"},
                {"projection_plane": "xz"},
                {"projection_plane": "yz"},
            ],
            "modes": ["mean"],
        },
        "circularity": {
            "kwargs": [
                {"projection_plane": "xy"},
                {"projection_plane": "xz"},
                {"projection_plane": "yz"},
            ],
            "modes": ["mean"],
        },
        "shape_factor": {
            "kwargs": [
                {"projection_plane": "xy"},
                {"projection_plane": "xz"},
                {"projection_plane": "yz"},
            ],
            "modes": ["mean"],
        },
        "length_fraction_above_soma": {
            "kwargs": [{"up": "X"}, {"up": "Y"}, {"up": "Z"}],
            "modes": ["mean"],
        },
        "volume_density": {"modes": ["mean"]},
        "principal_direction_extents": {"modes": ["max"]},
        "total_height": {"modes": ["max"]},
        "total_width": {"modes": ["max"]},
        "total_depth": {"modes": ["max"]},
        "total_length": {"modes": ["sum"]},
        "total_area": {"modes": ["sum"]},
    },
    "morphology": {"soma_radius": {"modes": ["mean"]}, "soma_surface_area": {"modes": ["mean"]}},
    "neurite_type": ["AXON", "APICAL_DENDRITE", "BASAL_DENDRITE"],
}


def select_population(
    input_morphologies: str, region: Optional[str], mtype: Optional[str]
) -> neurom.core.Population:
    """Compute the placeholder values for a given region - mtype couple."""
    logger.debug("Get population for %s in %s from %s", mtype, region, input_morphologies)

    # Load the morphologies for a selected region - mtype couple
    input_dir = Path(input_morphologies)
    metadata_path = input_dir / "metadata.csv"
    if metadata_path.exists():
        metadata = pd.read_csv(metadata_path)
        if region is not None:
            region_mask = metadata["brain_region"] == region
        else:
            region_mask = np.ones(len(metadata), dtype=bool)
        if mtype is not None:
            mtype_mask = metadata["cell_type"] == mtype
        else:
            mtype_mask = np.ones(len(metadata), dtype=bool)
        metadata = metadata.loc[region_mask & mtype_mask]
        population = neurom.load_morphologies(input_dir / metadata["morphology"])
    else:
        warnings.warn("No metadata.csv file found in the input directory, loading all morphologies")
        population = neurom.load_morphologies(input_dir)

    return population


def compute_placeholders(
    input_morphologies: str,
    region: str = None,
    mtype: str = None,
    config: Optional[dict] = None,
) -> pd.DataFrame:
    """Compute the placeholder values for a given region - mtype couple."""
    # Select morphologies
    population = select_population(input_morphologies, region, mtype)

    # Extract dataframe
    if config is None:
        config = DEFAULT_CONFIG

    if len(population) == 0:
        logger.debug(
            (
                "The population for the %s region and %s mtype is empty so the default "
                "placeholders are used."
            ),
            region,
            mtype,
        )
        df_placeholder = pd.read_csv(
            resource_filename(
                "morphology_workflows",
                "_data/default_placeholders.csv",
            ),
            header=[0, 1],
        )
    else:
        logger.debug("Compute placeholders with the following config: %s", config)
        df_placeholder = morph_stats.extract_dataframe(population, config)

    # Add region and mtype to the dataframe
    df_placeholder[("Metadata", "Region")] = region
    df_placeholder[("Metadata", "Mtype")] = mtype

    return df_placeholder



"""Script to fetch morphologies used in examples."""
import json
import re
import ssl
from pathlib import Path
from urllib.request import Request
from urllib.request import urlopen

import pandas as pd

NEUROMORPHO_URL = "http://neuromorpho.org"


def get_url(url):
    """Open a URL without SSL verification."""
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    ctx.set_ciphers("DEFAULT@SECLEVEL=1")
    req = Request(url)
    response = urlopen(req, context=ctx)  # pylint: disable=consider-using-with
    return response


def get_swc_by_neuron_index(neuronIndex, folder="morphologies"):
    """Download a neuron by index and store it into a SWC file.

    Args:
        neuronIndex(int): the neuron index in the database.
        folder(str): The folder in which the morphologies are written.

    Adapted from https://github.com/NeuroBox3D/neuromorpho/blob/master/rest_wrapper/rest_wrapper.py
    """
    url = f"{NEUROMORPHO_URL}/api/neuron/id/{neuronIndex}"
    response = get_url(url)

    neuron_name = json.loads(response.read().decode("utf-8"))["neuron_name"]
    url = f"{NEUROMORPHO_URL}/neuron_info.jsp?neuron_name={neuron_name}"
    html = get_url(url).read().decode("utf-8")
    p = re.compile(r"<a href=dableFiles/(.*)>Morphology File \(Standardized\)</a>", re.MULTILINE)
    m = re.findall(p, html)
    for match in m:
        file_name = match.replace("%20", " ").split("/")[-1]
        response = get_url(f"{NEUROMORPHO_URL}/dableFiles/{match}")
        filename = folder / file_name
        with open(filename, "w", encoding="utf-8") as f:
            f.write(response.read().decode("utf-8"))
        print("Fetched", filename)
        return filename


def main():
    """Select neurons and download them from the NeuroMorpho database."""
    brainRegion = "barrel"
    numNeurons = 500

    url = f"{NEUROMORPHO_URL}/api/neuron/select?q=brain_region:{brainRegion}&size={numNeurons}"
    response = get_url(url)

    neurons = json.loads(response.read().decode("utf-8"))
    df = pd.DataFrame(neurons["_embedded"]["neuronResources"])
    df["type"] = df["cell_type"].apply(lambda t: t[-1])
    df = df[df["type"] == "principal cell"]
    folder = Path("morphologies")
    folder.mkdir(exist_ok=True)
    for gid in df.index:
        filename = get_swc_by_neuron_index(df.loc[gid, "neuron_id"], folder=folder)
        df.loc[gid, "morph_path"] = filename

    df["morph_name"] = df["neuron_name"]
    df[["morph_path", "morph_name", "brain_region"]].to_csv("dataset.csv")


if __name__ == "__main__":
    main()



"""CLI for validation workflows."""
import argparse
import inspect
import logging
import os
import re
import sys
from pathlib import Path

import luigi
from luigi_tools.util import get_dependency_graph
from luigi_tools.util import graphviz_dependency_graph
from luigi_tools.util import render_dependency_graph

import morphology_workflows
from morphology_workflows.tasks import workflows
from morphology_workflows.tasks.fetch import Fetch
from morphology_workflows.tasks.placeholders import Placeholders

L = logging.getLogger(__name__)


WORKFLOW_TASKS = {
    "Fetch": Fetch,
    "Placeholders": Placeholders,
    "Curate": workflows.Curate,
    "Annotate": workflows.Annotate,
    "Repair": workflows.Repair,
}

LOGGING_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

LUIGI_PARAMETERS = ["workers", "log_level"]


_PARAM_NO_VALUE = [luigi.parameter._no_value, None]  # pylint: disable=protected-access


def _process_param(param):
    desc = param.description
    choices = None
    interval = None
    optional = False
    if isinstance(param, luigi.OptionalParameter):
        optional = True
    if isinstance(param, luigi.ChoiceParameter):
        desc, choices = desc.rsplit("Choices: ", 1)
    if isinstance(param, luigi.NumericalParameter):
        desc, interval = desc.rsplit("permitted values: ", 1)
    try:
        param_type, param_doc = re.match("(:.*?:)? *(.*)", desc).groups()
    except AttributeError:
        param_type = None
        param_doc = desc
    return param_doc, param_type, choices, interval, optional


def format_description(
    param,
    default_str="{doc} Default value: {default}.",
    optional_str="(optional) {doc}",
    type_str="({type}) {doc}",
    choices_str="{doc} Choices: {choices}.",
    interval_str="{doc} Permitted values: {interval}.",
    param_no_value=None,
):
    """Format the description of a parameter."""
    if param_no_value is None:
        param_no_value = _PARAM_NO_VALUE

    try:
        param_doc, param_type, choices, interval, optional = _process_param(param)
        if optional:
            param_doc = optional_str + param_doc
        if param_type is not None:
            param_doc = type_str.format(doc=param_doc, type=param_type.replace(":", ""))
        if choices is not None:
            param_doc = choices_str.format(doc=param_doc, choices=choices)
        if interval is not None:
            param_doc = interval_str.format(doc=param_doc, interval=interval)
        # pylint: disable=protected-access
        if hasattr(param, "_default") and param._default not in param_no_value:
            param_doc = default_str.format(doc=param_doc, default=param._default)
    except AttributeError:
        param_doc = param.description
    return param_doc


class ArgParser:
    """Class to build parser and parse arguments."""

    def __init__(self):
        self.parsers = self._get_parsers()

    @property
    def parser(self):
        """Return the root parser."""
        return self.parsers["root"]

    def _get_parsers(self):
        """Return the main argument parser."""
        parser = argparse.ArgumentParser(
            description="Run the workflow",
        )

        parser.add_argument("-c", "--config-path", help="Path to the Luigi config file")

        parser.add_argument(
            "-m",
            "--master-scheduler",
            default=False,
            action="store_true",
            help="Use Luigi's master scheduler instead of local scheduler.",
        )

        parser.add_argument(
            "-ll",
            "--log-level",
            default="INFO",
            choices=LOGGING_LEVELS,
            help="Logger level.",
        )

        parser.add_argument("-lf", "--log-file", help="Logger file.")

        parser.add_argument(
            "-w",
            "--workers",
            type=int,
            default=1,
            help="Number of workers that luigi can summon.",
        )

        parser.add_argument(
            "-dg",
            "--create-dependency-graph",
            help=(
                "Create the dependency graph of a workflow instead of running it. "
                "Pass a path to render it as an image (depending on the extension of the given "
                "path)."
            ),
        )

        return self._get_workflow_parsers(parser)

    @staticmethod
    def _get_workflow_parsers(parser=None):
        """Return the workflow argument parser.

        If parser is None, a new parser is created with the workflows as subparsers,
        otherwise if it is supplied, the parsers are added as subparsers.

        For each task listed in WORKFLOW_TASKS, a subparser is created as if it was
        created by luigi.
        """
        if not parser:
            parser = argparse.ArgumentParser()

        parsers = {"root": parser}

        workflow_parser = parser.add_subparsers(help="Possible workflows", dest="workflow")

        for workflow_name, task in WORKFLOW_TASKS.items():
            try:
                task_name = task.__name__
                doc = task.__doc__
                if ".. graphviz::" in doc:
                    doc = re.sub(
                        (
                            "The complete phase has the following dependency graph:"
                            r"\n\n    .. graphviz:: .*\.dot"
                        ),
                        "",
                        doc,
                        flags=re.DOTALL,
                    ).strip()
                subparser = workflow_parser.add_parser(workflow_name, help=doc)
                for param, param_obj in task.get_params():
                    param_name = "--" + param.replace("_", "-")
                    subparser.add_argument(
                        param_name,
                        help=format_description(param_obj),
                        # pylint: disable=protected-access
                        **param_obj._parser_kwargs(param, task_name),
                    )
                parsers[workflow_name] = subparser
            except (AttributeError, TypeError):
                pass

        return parsers

    def parse_args(self, argv):
        """Parse the arguments, and return a argparse.Namespace object."""
        args = self.parser.parse_args(argv)

        return args


def _build_parser():
    """Build the parser."""
    tmp = ArgParser().parser
    return tmp


def main(arguments=None):
    """Main function."""
    logging.getLogger("luigi-interface").propagate = False

    if arguments is None:
        arguments = sys.argv[1:]

    # Parse arguments
    parser = ArgParser()
    args = parser.parse_args(arguments)

    L.debug("Args: %s", args)

    # Check that one workflow is in arguments
    if args is None or args.workflow is None:
        L.critical("Arguments must contain one workflow. Check help with -h/--help argument.")
        parser.parser.print_help()
        sys.exit()

    # Set luigi.cfg path
    if args.config_path is not None:
        os.environ["LUIGI_CONFIG_PATH"] = args.config_path

    # Get arguments to configure luigi
    luigi_config = {k: v for k, v in vars(args).items() if k in LUIGI_PARAMETERS}
    luigi_config["local_scheduler"] = not args.master_scheduler

    # Prepare workflow task and arguments
    task_cls = WORKFLOW_TASKS[args.workflow]
    args_dict = {k.split(task_cls.get_task_family() + "_")[-1]: v for k, v in vars(args).items()}
    task_params = [i for i, j in task_cls.get_params()]
    args_dict = {k: v for k, v in args_dict.items() if v is not None and k in task_params}
    task = WORKFLOW_TASKS[args.workflow](**args_dict)

    # Export the dependency graph of the workflow instead of running it
    if args.create_dependency_graph is not None:
        g = get_dependency_graph(task, allow_orphans=True)

        # Create URLs
        base_f = Path(inspect.getfile(morphology_workflows)).parent
        node_kwargs = {}
        for _, child in g:
            if child is None:
                continue
            url = (
                Path(inspect.getfile(child.__class__)).relative_to(base_f).with_suffix("")
                / "index.html"
            )
            anchor = "#" + ".".join(child.__module__.split(".")[1:] + [child.__class__.__name__])
            node_kwargs[child] = {"URL": "../../" + url.as_posix() + anchor}

        dot = graphviz_dependency_graph(g, node_kwargs=node_kwargs)
        render_dependency_graph(dot, args.create_dependency_graph)
        return

    # Run the luigi task
    luigi.build([task], **luigi_config)


if __name__ == "__main__":
    main()



"""Annotation tasks."""
import logging

import luigi
from data_validation_framework.task import ElementValidationTask
from data_validation_framework.task import SkippableMixin
from luigi_tools.parameter import BoolParameter

from morphology_workflows.annotation import assign_mtypes
from morphology_workflows.annotation import compute_hard_limits
from morphology_workflows.annotation import detect_cut_leaves
from morphology_workflows.annotation import find_apical_point
from morphology_workflows.annotation import plot_apical_point
from morphology_workflows.annotation import plot_cut_leaves
from morphology_workflows.annotation import plot_hard_limits
from morphology_workflows.curation import collect

logger = logging.getLogger(__name__)


class CollectCurated(ElementValidationTask):
    """Collect curated dataset to work with on this phase."""

    output_columns = {
        "morph_path": None,
        "has_apical": None,
        "mtype": None,
        "rotation_matrix": None,
    }

    validation_function = collect


class MType(ElementValidationTask):
    """Assign mtypes to morphologies.

    Currently, we only check if the mtype corresponds to the provided list of regex.
    If the list of regex is empty, we let all the morphologies pass (equivalent to skip).

    If they do not, we invalidate the morphologies.
    """

    output_columns = {"mtype": None}
    validation_function = assign_mtypes
    mtype_regex = luigi.OptionalListParameter(default=None)

    def kwargs(self):
        """ """
        return {"mtype_regex": self.mtype_regex}

    def inputs(self):
        """ """
        return {CollectCurated: {"mtype": "mtype"}}


class HardLimit(ElementValidationTask):
    """Compute hard limits.

    Hard limits are markers used to place morphologies in a circuit. They can be automatically
    computed as min and max extent of axon and dendrites.

    .. todo:: There can be more such markers, manually placed with ``nse-tools/placement-annotator``
        internal tool.
    """

    output_columns = {"hard_limit_path": None}
    validation_function = compute_hard_limits

    skip_axon = BoolParameter(
        default=False, description=":bool: Skip the computation for axon hard limit"
    )
    dendrite_hard_limit = luigi.Parameter(
        default="L1_hard_limit", description=":str: Name of the dendrite hard limit annotation"
    )
    axon_hard_limit = luigi.Parameter(
        default="L1_axon_hard_limit", description=":str: Name of the axon hard limit annotation"
    )

    def kwargs(self):
        """ """
        return {
            "skip_axon": self.skip_axon,
            "dendrite_hard_limit": self.dendrite_hard_limit,
            "axon_hard_limit": self.axon_hard_limit,
        }

    def inputs(self):
        """ """
        return {CollectCurated: {"morph_path": "morph_path"}}


class PlotHardLimit(SkippableMixin(), ElementValidationTask):
    """Plot the hard limits.

    Plot hard limits as lines on morphologies with plotly.
    """

    output_columns = {"plot_hard_limit_path": None}
    validation_function = plot_hard_limits

    with_plotly = BoolParameter(default=True, description=":bool: Use Plotly for plotting")

    def kwargs(self):
        """ """
        return {"with_plotly": self.with_plotly}

    def inputs(self):
        """ """
        return {
            CollectCurated: {"morph_path": "morph_path"},
            HardLimit: {"hard_limit_path": "hard_limit_path"},
        }


class ApicalPoint(SkippableMixin(), ElementValidationTask):
    """Detect apical point.

    For morphologies with apical dendrites we estimate the location of the apical point with
    the automatic tool :func:`morph_tool.apical_point.apical_point_position`.
    """

    output_columns = {"apical_point_path": None}
    validation_function = find_apical_point

    tuft_percent = luigi.FloatParameter(
        default=20, description=":float: Fraction of tuft for morph-tool algorithm"
    )

    def kwargs(self):
        """ """
        return {"tuft_percent": self.tuft_percent}

    def inputs(self):
        """ """
        return {CollectCurated: {"morph_path": "morph_path", "has_apical": "has_apical"}}


class PlotApicalPoint(SkippableMixin(), ElementValidationTask):
    """Plot apical point.

    Plot apical point as a single scatter point on a morphology with plotly.
    """

    output_columns = {"plot_apical_point_path": None}
    validation_function = plot_apical_point

    with_plotly = BoolParameter(default=True, description=":bool: Use Plotly for plotting")

    def kwargs(self):
        """ """
        return {"with_plotly": self.with_plotly}

    def inputs(self):
        """ """
        return {
            CollectCurated: {"morph_path": "morph_path", "has_apical": "has_apical"},
            ApicalPoint: {"apical_point_path": "apical_point_path"},
        }


class CutLeaves(ElementValidationTask):
    """Detect the cut leaves.

    Cut leaves are considered to represent boundaries of slices of in-vitro reconstructions.
    See :func:`neuror.cut_plane.cut_leaves` for more details on the algorithm and its parameters.
    """

    output_columns = {"cut_leaves_path": None, "cut_qualities": None}
    validation_function = detect_cut_leaves

    bin_width = luigi.FloatParameter(
        default=15, description=":float: Thickness of cut plane, (used if legacy_detection=False)"
    )
    percentile_threshold = luigi.FloatParameter(
        default=70, description=":float: Threshold percenetile for finding a cut plane"
    )

    def kwargs(self):
        """ """
        return {"bin_width": self.bin_width, "percentile_threshold": self.percentile_threshold}

    def inputs(self):
        """ """
        return {CollectCurated: {"morph_path": "morph_path", "rotation_matrix": "rotation_matrix"}}


class PlotCutLeaves(SkippableMixin(), ElementValidationTask):
    """Plot the cut leaves.

    We plot the cut-leaves as scatter markers, as the terminal points on dendrites intersecting
    the cut plane.
    """

    output_columns = {"plot_cut_leaves_path": None}
    validation_function = plot_cut_leaves

    with_plotly = BoolParameter(default=True, description=":bool: Use Plotly for plotting")

    def kwargs(self):
        """ """
        return {"with_plotly": self.with_plotly}

    def inputs(self):
        """ """
        return {
            CutLeaves: {"cut_leaves_path": "cut_leaves_path", "cut_qualities": "cut_qualities"},
            CollectCurated: {"morph_path": "morph_path"},
        }



"""Configuration file for the Sphinx documentation builder."""

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import importlib
import os
import re
import subprocess
from importlib import metadata
from pathlib import Path

import luigi
from sphinx.util import logging

import morphology_workflows
import morphology_workflows.tasks
from morphology_workflows.tasks import cli

logger = logging.getLogger(__name__)

# -- Project information -----------------------------------------------------

project = "morphology-workflows"

# The short X.Y version
version = None
if os.environ.get("READTHEDOCS_VERSION", "") == "stable":
    logger.info("Stable ReadTheDocs environment found.")
    os.environ["SPHINX_BLUE_BRAIN_THEME_CHECK_VERSIONS"] = "False"
    with subprocess.Popen(
        ["git", "describe", "--abbrev=0", "--tags"],
        stdout=subprocess.PIPE,
    ) as proc:
        version = proc.communicate()[0].decode("utf-8").strip()
    if not version:
        logger.info(
            "Could not find version from the 'git describe' command ."
            "The 'importlib.metadata.version' function will be used."
        )

if not version:
    version = metadata.version("morphology-workflows")

logger.info(f"Version found: {version}")

# The full version, including alpha/beta/rc tags
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "autoapi.extension",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinxarg.ext",
    "m2r2",
]

autoapi_dirs = [
    "../../src/morphology_workflows",
    "../../src/morphology_workflows/tasks",
]
autoapi_ignore = [
    "*_templates/**",
]
autoapi_python_use_implicit_namespaces = True
autoapi_keep_files = False
autoapi_add_toctree_entry = False
autoapi_options = [
    "imported-members",
    "members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "undoc-members",
]
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx-bluebrain-theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

html_theme_options = {
    "metadata_distribution": "morphology-workflows",
}

html_title = "Morphology Workflows"

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# autosummary settings
autosummary_generate = True

# autodoc settings
autodoc_typehints = "signature"
autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
}

intersphinx_mapping = {
    # TODO: Uncomment this once it is published on RTD
    # "data-validation-framework": (
    #     "https://data-validation-framework.readthedocs.io/en/stable",
    #     None,
    # ),
    "diameter-synthesis": ("https://diameter-synthesis.readthedocs.io/en/stable", None),
    "luigi": ("https://luigi.readthedocs.io/en/stable", None),
    "luigi-tools": ("https://luigi-tools.readthedocs.io/en/stable", None),
    "neurom": ("https://neurom.readthedocs.io/en/stable", None),
    "neuror": ("https://neuror.readthedocs.io/en/stable", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs", None),
    "python": ("https://docs.python.org/3", None),
}

SKIP = [
    r".*\.L$",
    r".*tasks\..*\.__specifications__$",
    r".*tasks\..*\.args$",
    r".*tasks\..*\.input_index_col$",
    r".*tasks\..*\.inputs$",
    r".*tasks\..*\.kwargs$",
    r".*tasks\..*\.output$",
    r".*tasks\..*\.output_columns$",
    r".*tasks\..*\.requires$",
    r".*tasks\..*\.run$",
    r".*tasks\..*\.validation_function$",
]

IMPORT_MAPPING = {
    "morphology_workflows": morphology_workflows,
    "tasks": morphology_workflows.tasks,
}

_PARAM_NO_VALUE = [luigi.parameter._no_value, None]  # pylint: disable=protected-access


# pylint: disable=unused-argument
def maybe_skip_member(app, what, name, obj, skip, options):
    """Skip useless members and format documentation of others."""
    skip = None
    for pattern in SKIP:
        if re.match(pattern, name) is not None:
            skip = True
            break

    if not skip:
        if name.count(".") < 3:
            return skip
        package, module, *path = name.split(".")
        root_package = IMPORT_MAPPING[package]
        actual_module = importlib.import_module(root_package.__name__ + "." + module)

        try:
            task = getattr(actual_module, path[-2])
            actual_obj = getattr(task, path[-1])
        except AttributeError:
            return skip

        if isinstance(actual_obj, luigi.Parameter):
            if hasattr(actual_obj, "description") and actual_obj.description:
                obj.docstring = cli.format_description(
                    actual_obj,
                    default_str="{doc}\n\n:default value: {default}",
                    optional_str="(optional) {doc}",
                    type_str="{doc}\n\n:type: {type}",
                    choices_str="{doc}\n\n:choices: {choices}",
                    interval_str="{doc}\n\n:permitted values: {interval}",
                )

    return skip


def generate_images(*args, **kwargs):
    """Generate images of the workflows."""
    old_cwd = os.getcwd()
    try:
        os.environ["LUIGI_CONFIG_PATH"] = str(
            Path(*Path(__file__).parts[:-3]) / "tests/data/test_example_1/luigi.cfg"
        )
        cur_cwd = Path(__file__).parent
        cli.main(["-dg", str(cur_cwd / "autoapi/tasks/workflows/Curate.dot"), "Curate"])
        cli.main(["-dg", str(cur_cwd / "autoapi/tasks/workflows/Curate.png"), "Curate"])
        cli.main(["-dg", str(cur_cwd / "autoapi/tasks/workflows/Annotate.dot"), "Annotate"])
        cli.main(["-dg", str(cur_cwd / "autoapi/tasks/workflows/Annotate.png"), "Annotate"])
        cli.main(["-dg", str(cur_cwd / "autoapi/tasks/workflows/Repair.dot"), "Repair"])
        cli.main(["-dg", str(cur_cwd / "autoapi/tasks/workflows/Repair.png"), "Repair"])
    finally:
        os.chdir(old_cwd)


def setup(app):
    """Setup Sphinx by connecting functions to events."""
    app.connect("builder-inited", generate_images)
    app.connect("autoapi-skip-member", maybe_skip_member)



"""Curation tasks."""
import logging

import luigi
from data_validation_framework.task import ElementValidationTask
from data_validation_framework.task import SetValidationTask
from data_validation_framework.task import SkippableMixin
from luigi.parameter import OptionalChoiceParameter
from luigi.parameter import OptionalNumericalParameter
from luigi_tools.parameter import BoolParameter
from neuror.sanitize import _ZERO_LENGTH

from morphology_workflows.curation import align
from morphology_workflows.curation import check_neurites
from morphology_workflows.curation import collect
from morphology_workflows.curation import detect_errors
from morphology_workflows.curation import extract_marker
from morphology_workflows.curation import fix_neurites_in_soma
from morphology_workflows.curation import make_error_report
from morphology_workflows.curation import orient
from morphology_workflows.curation import plot_errors
from morphology_workflows.curation import plot_markers
from morphology_workflows.curation import plot_morphology
from morphology_workflows.curation import recenter
from morphology_workflows.curation import resample
from morphology_workflows.curation import sanitize

logger = logging.getLogger(__name__)


class Collect(ElementValidationTask):
    """Collect external dataset from .csv file.

    Original dataset has to have a 'morph_name' and 'morph_path' columns, with the name and path to
    the corresponding morphologies. Any other valid columns will be gathered as well.
    In addition, it will only collect morphologies with allowed extension: ``.asc``, ``.h5``,
    ``.swc``.
    """

    output_columns = {"morph_path": None}
    validation_function = collect


class ExtractMarkers(SkippableMixin(), ElementValidationTask):
    """Extract marker information from the original morphology files, if any.

    Markers are additional spatial information contained in some reconstructed morphologies.
    Some can be read by MorphIO (https://github.com/BlueBrain/MorphIO/pull/186).
    The markers are stored in a custom API: :mod:`morphology_workflows.marker_helper`.
    """

    output_columns = {"marker_path": None}
    validation_function = extract_marker

    def inputs(self):
        """ """
        return {Collect: {"morph_path": "morph_path"}}


class PlotMarkers(SkippableMixin(), ElementValidationTask):
    """Plot markers on morphologies.

    Plot the markers extracted from :class:`tasks.curation.ExtractMarkers` on the morphologies.
    """

    output_columns = {"plot_marker_path": None}
    validation_function = plot_markers

    with_plotly = BoolParameter(default=True, description=":bool: Use Plotly for plotting")

    def kwargs(self):
        """ """
        return {"with_plotly": self.with_plotly}

    def inputs(self):
        """ """
        return {
            ExtractMarkers: {"marker_path": "marker_path"},
        }


class CheckNeurites(ElementValidationTask):
    """Detect which neurites are present in the morphology, and add soma if missing.

    This task adds three important boolean flags:
        - has_axon
        - has_basal
        - has_apical

    set to ``False`` if any are absent.

    For the axon, if it only has fewer sections than ``axon_n_section_min``, it will be considered
    as ``has_axon=False``.

    If soma is absent, a soma is added as a circle, with center the mean of the first points of root
    sections, and radius as standard deviation of these points around the center.

    If the input dataset has columns ``use_axon`` or ``use_dendrites``, our corresponding flags
    will match.

    .. todo::
        Set various levels, a one section neurite can be set to ``has_basal=False`` for
        example.
    """

    output_columns = {
        "morph_path": None,
        "has_axon": False,
        "has_basal": False,
        "has_apical": False,
    }
    validation_function = check_neurites

    mock_soma_type = OptionalChoiceParameter(
        description=":str: Add a soma if missing of the given type",
        choices=["spherical", "contour"],
        default="spherical",
    )
    axon_n_section_min = luigi.IntParameter(
        description=":int: Minimum number of sections in an axon to consider it valid",
        default=5,
    )
    ensure_stub_axon = luigi.BoolParameter(
        description=":bool: Add a stub axon if there is no axon on the morphology",
        default=False,
    )
    min_length_first_section = OptionalNumericalParameter(
        description=(
            ":float: Resize the first section to be at least of the given size (do nothing if None "
            "is given)"
        ),
        default=_ZERO_LENGTH,
        var_type=float,
        min_value=0,
        max_value=float("inf"),
        left_op=luigi.parameter.operator.lt,
    )

    def kwargs(self):
        """ """
        return {
            "mock_soma_type": self.mock_soma_type,
            "axon_n_section_min": self.axon_n_section_min,
            "ensure_stub_axon": self.ensure_stub_axon,
            "min_length_first_section": self.min_length_first_section,
        }

    def inputs(self):
        """ """
        return {Collect: {"morph_path": "morph_path"}}


class Sanitize(ElementValidationTask):
    """Sanitize the morphologies.

    Sanitization is done with :func:`neuror.sanitize.sanitize` and does the following:

    - ensures it can be loaded with MorphIO
    - raises if the morphology has no soma or of invalid format
    - removes unifurcations
    - set negative diameters to zero
    - raises if the morphology has a neurite whose type changes along the way
    - removes segments with near zero lengths (shorter than 1e-4)

    Note that the :class:`tasks.curation.CheckNeurite` task adds a soma if missing, so a failure
    here means that the soma does not have a valid type.
    """

    ensure_roots_at_soma = luigi.BoolParameter(
        default=True, description=":bool: Reconnect root sections at soma if they are not"
    )
    output_columns = {"morph_path": None}

    validation_function = sanitize

    def kwargs(self):
        """ """
        return {"ensure_roots_at_soma": self.ensure_roots_at_soma}

    def inputs(self):
        """ """
        return {CheckNeurites: {"morph_path": "morph_path"}}


class Recenter(SkippableMixin(), ElementValidationTask):
    """Recenter morphologies.

    Often, morphologies do not have a soma centered at ``[0, 0, 0]``, so we recenter and save
    the original location, in case it is important to know where the morphology is located in atlas.
    """

    output_columns = {"morph_path": None, "soma_location": ""}
    validation_function = recenter

    def inputs(self):
        """ """
        return {Sanitize: {"morph_path": "morph_path"}}


class Orient(ElementValidationTask):
    """Orient morphologies.

    Sometimes, morphologies are oriented along non-standard axis. At BBP, the standard axis is
        - ``y`` for pia direction (apical trunk, minus main axon, etc...)
        - ``z`` is the slice direction is invitro
        - ``x`` the other direction

    If ``y`` is provided (default value), no oriention will be applied.
    """

    output_columns = {"morph_path": None}
    validation_function = orient

    pia_direction = luigi.Parameter(
        default="y", description=":str: Axis of pia direction, x, y or z"
    )

    def kwargs(self):
        """ """
        return {"pia_direction": self.pia_direction}

    def inputs(self):
        """ """
        return {Recenter: {"morph_path": "morph_path"}}


class Align(SkippableMixin(True), ElementValidationTask):
    """Align morphologies.

    Sometimes, a morphology is not aligned with any consistent direction, so we can try here to
    align them using a various algorithms from :func:`morph_tool.transform.align_morphology`.

    ``method`` can be: ``whole``, ``trunk``, ``first_segment``, ``first_section``, ``custom``

    If ``method`` is ``custom``, one must provide external orientation direction in a json file
    via the parameter ``custom_orientation_json_path``.


    TODO: implement the axon_collateral method (for thalamus or example), following
    https://bbpgitlab.epfl.ch/cells/archive/iavarone/thalamus_pipeline/-/blob/master/notebooks/cell_alignement/Rt_TC_rotations_AxonCollaterals_Releases2019.ipynb

    By default, this task is skipped.
    """

    output_columns = {"morph_path": None, "rotation_matrix": ""}
    validation_function = align

    method = luigi.ChoiceParameter(
        default="whole",
        choices=["whole", "trunk", "first_segment", "first_section", "custom"],
        description=":str: Method to align morphology",
    )
    neurite_type = luigi.Parameter(
        default="apical", description=":str: Neurite to use to align morphology"
    )
    direction = luigi.OptionalListParameter(default=None)
    custom_orientation_json_path = luigi.OptionalStrParameter(
        default=None, description=":str: Path to json with custom orientations"
    )

    def kwargs(self):
        """ """
        return {
            "method": self.method,
            "neurite_type": self.neurite_type,
            "direction": self.direction,
            "custom_orientation_json_path": self.custom_orientation_json_path,
        }

    def inputs(self):
        """ """
        return {Orient: {"morph_path": "morph_path"}}


class EnsureNeuritesOutsideSoma(SkippableMixin(True), ElementValidationTask):
    """Fix radius of the soma of each morphology and cut the root section of neurites if needed."""

    output_columns = {"morph_path": None}
    validation_function = fix_neurites_in_soma

    def inputs(self):
        """ """
        return {Align: {"morph_path": "morph_path"}}


class DetectErrors(SkippableMixin(), ElementValidationTask):
    """Detect errors in reconstructions.

    Reconstructions may contain errors, which are detected here.
    They are of the following type:

    - fat ends
    - z-jumps
    - narrow start
    - dangling branch
    - multifurcation
    - z-range (new error only present here, which check is the z thickness is larger than min_range)

    This task uses NeuroR/neuror/error_annotation.py (https://github.com/BlueBrain/NeuroR),
    and reproduuces part of what is in MorphService.
    This task creates new .asc file with error annotated so it can be red by Neuroluscida,
    and a MarkerSet container of the errors, for later plotting.
    """

    output_columns = {
        "error_marker_path": None,
        "error_annotated_path": None,
        "error_summary": None,
    }
    validation_function = detect_errors

    min_range = luigi.FloatParameter(
        default=50, description=":float: Minimum z-range to be an error"
    )

    def kwargs(self):
        """ """
        return {"min_range": self.min_range}

    def inputs(self):
        """ """
        return {CheckNeurites: {"morph_path": "morph_path"}}


class PlotErrors(SkippableMixin(), ElementValidationTask):
    """Plot detected errors.

    From the detected errors in :class:`tasks.curation.DetectErrors`, plot them on the morphologies.
    """

    output_columns = {"plot_errors_path": ""}
    validation_function = plot_errors

    with_plotly = BoolParameter(default=True, description=":bool: Use Plotly for plotting")

    def kwargs(self):
        """ """
        return {"with_plotly": self.with_plotly}

    def inputs(self):
        """ """
        return {
            Recenter: {"morph_path": "morph_path"},
            DetectErrors: {"error_marker_path": "error_marker_path"},
        }


class ErrorsReport(SkippableMixin(), SetValidationTask):
    """Save error report for all morphologies."""

    error_report_path = luigi.Parameter(
        default="error_report.csv", description=":str: Path to error report file in .csv"
    )
    validation_function = make_error_report

    def kwargs(self):
        """ """
        return {"error_report_path": self.error_report_path}

    def inputs(self):
        """ """
        return {
            DetectErrors: {"error_marker_path": "error_marker_path"},
        }


class Resample(SkippableMixin(), ElementValidationTask):
    """Resample morphologies.

    This tasks ensures a constant sampling rate of points along all branches.
    """

    output_columns = {"morph_path": None}
    validation_function = resample

    linear_density = luigi.FloatParameter(
        default=1.0, description=":float: Density of points per micron"
    )

    def kwargs(self):
        """ """
        return {"linear_density": self.linear_density}

    def inputs(self):
        """ """
        return {EnsureNeuritesOutsideSoma: {"morph_path": "morph_path"}}


class PlotMorphologies(SkippableMixin(), ElementValidationTask):
    """Plot curated morphologies.

    This tasks creates a single pdf with all morphologies to visualise them after curation.
    """

    output_columns = {"plot_path": None}
    validation_function = plot_morphology

    with_plotly = BoolParameter(default=False, description=":bool: Use Plotly for plotting")
    with_realistic_diameters = BoolParameter(
        default=True,
        description=":bool: Use realistic diameters for plotting (take more time to plot)",
    )

    def kwargs(self):
        """ """
        return {
            "with_plotly": self.with_plotly,
            "realistic_diameters": self.with_realistic_diameters,
        }

    def inputs(self):
        """ """
        return {
            Resample: {"morph_path": "morph_path"},
        }



"""Test the complete workflow on the example."""
# pylint: disable=redefined-outer-name
import shutil

import luigi
import numpy as np
import pytest
from dir_content_diff import assert_equal_trees

from morphology_workflows.tasks import workflows

from . import clean_exception


@pytest.fixture()
def example_1(tmp_working_dir, examples_dir):
    """Setup the working directory."""
    shutil.copyfile(examples_dir / "logging.conf", tmp_working_dir / "logging.conf")
    shutil.copyfile(examples_dir / "luigi.cfg", tmp_working_dir / "luigi.cfg")
    shutil.copyfile(examples_dir / "dataset.csv", tmp_working_dir / "dataset.csv")
    shutil.copyfile(examples_dir / "builder_recipe.xml", tmp_working_dir / "builder_recipe.xml")
    shutil.copyfile(examples_dir / "placement_rules.xml", tmp_working_dir / "placement_rules.xml")
    shutil.copyfile(examples_dir / "transform_rules.xml", tmp_working_dir / "transform_rules.xml")
    shutil.copytree(examples_dir / "morphologies", tmp_working_dir / "morphologies")

    # Set current config in luigi
    luigi_config = luigi.configuration.get_config()
    luigi_config.read("./luigi.cfg")

    yield tmp_working_dir

    # Reset luigi config
    luigi_config.clear()


def test_example_1(example_1, data_dir):
    """Test the workflow on the example."""
    np.random.seed(0)

    # Run the Curate workflow
    assert luigi.build([workflows.Curate()], local_scheduler=True)

    # Define post-processes to check the results
    out_dir_pattern = (str(example_1) + "/?", "")
    curation_expected_dir = data_dir / "test_example_1" / "out_curated"
    curated_specific_args = {
        "curated_dataset.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "error_annotated_path",
                        "error_marker_path",
                        "marker_path",
                        "morph_path",
                    ]
                }
            }
        },
        "Curate/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "('Align', 'morph_path')",
                        "('CheckNeurites', 'morph_path')",
                        "('Collect', 'morph_path')",
                        "('DetectErrors', 'error_annotated_path')",
                        "('DetectErrors', 'error_marker_path')",
                        "('DetectErrors', 'morph_path')",
                        "('ExtractMarkers', 'marker_path')",
                        "('ExtractMarkers', 'morph_path')",
                        "('Orient', 'morph_path')",
                        "('EnsureNeuritesOutsideSoma', 'morph_path')",
                        "('PlotErrors', 'error_marker_path')",
                        "('PlotErrors', 'morph_path')",
                        "('PlotErrors', 'plot_errors_path')",
                        "('PlotMarkers', 'marker_path')",
                        "('PlotMarkers', 'plot_marker_path')",
                        "('PlotMorphologies', 'morph_path')",
                        "('PlotMorphologies', 'plot_path')",
                        "('Recenter', 'morph_path')",
                        "('Resample', 'morph_path')",
                        "('Sanitize', 'comment')",
                        "('Sanitize', 'morph_path')",
                        "('ErrorsReport', 'error_marker_path')",
                        "error_annotated_path",
                        "error_marker_path",
                        "marker_path",
                        "morph_path",
                    ],
                },
            }
        },
        "Collect/report.csv": {
            "format_data_kwargs": {"replace_pattern": {out_dir_pattern: ["morph_path"]}}
        },
        "CheckNeurites/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: ["morph_path"],
                },
            }
        },
        "ExtractMarkers/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: ["morph_path", "marker_path"],
                },
            }
        },
        "PlotMarkers/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: ["marker_path", "plot_marker_path"],
                }
            }
        },
        "Recenter/report.csv": {
            "format_data_kwargs": {"replace_pattern": {out_dir_pattern: ["morph_path"]}}
        },
        "Orient/report.csv": {
            "format_data_kwargs": {"replace_pattern": {out_dir_pattern: ["morph_path"]}}
        },
        "Align/report.csv": {
            "format_data_kwargs": {"replace_pattern": {out_dir_pattern: ["morph_path"]}}
        },
        "EnsureNeuritesOutsideSoma/report.csv": {
            "format_data_kwargs": {"replace_pattern": {out_dir_pattern: ["morph_path"]}}
        },
        "DetectErrors/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "morph_path",
                        "error_annotated_path",
                        "error_marker_path",
                    ]
                }
            }
        },
        "PlotErrors/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: ["morph_path", "error_marker_path", "plot_errors_path"]
                }
            }
        },
        "Sanitize/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {out_dir_pattern: ["morph_path", "comment"]},
            }
        },
        "PlotMorphologies/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {out_dir_pattern: ["morph_path", "plot_path"]}
            }
        },
    }

    # Check the results
    assert_equal_trees(
        curation_expected_dir, example_1 / "out_curated", specific_args=curated_specific_args
    )

    # Run the Annotate workflow
    assert luigi.build([workflows.Annotate()], local_scheduler=True)

    # Define post-processes to check the results
    annotation_expected_dir = data_dir / "test_example_1" / "out_annotated"
    annotated_specific_args = {
        "annotated_dataset.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "apical_point_path",
                        "cut_leaves_path",
                        "error_annotated_path",
                        "error_marker_path",
                        "hard_limit_path",
                        "marker_path",
                        "morph_path",
                    ]
                }
            }
        },
        "Annotate/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "('ApicalPoint', 'apical_point_path')",
                        "('ApicalPoint', 'morph_path')",
                        "('CollectCurated', 'morph_path')",
                        "('CutLeaves', 'cut_leaves_path')",
                        "('CutLeaves', 'morph_path')",
                        "('HardLimit', 'hard_limit_path')",
                        "('HardLimit', 'morph_path')",
                        "('PlotApicalPoint', 'apical_point_path')",
                        "('PlotApicalPoint', 'morph_path')",
                        "('PlotApicalPoint', 'plot_apical_point_path')",
                        "('PlotCutLeaves', 'cut_leaves_path')",
                        "('PlotCutLeaves', 'morph_path')",
                        "('PlotCutLeaves', 'plot_cut_leaves_path')",
                        "('PlotHardLimit', 'hard_limit_path')",
                        "('PlotHardLimit', 'morph_path')",
                        "('PlotHardLimit', 'plot_hard_limit_path')",
                        "apical_point_path",
                        "cut_leaves_path",
                        "hard_limit_path",
                    ],
                    clean_exception(
                        "ValueError: Can not search for cut leaves for a neuron with no neurites"
                    ): ["('CutLeaves', 'exception')"],
                    clean_exception(
                        "ValueError: Can not plot axis marker for a neuron with no neurites"
                    ): ["('PlotHardLimit', 'exception')"],
                },
            }
        },
        "ApicalPoint/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "apical_point_path",
                        "morph_path",
                    ]
                }
            }
        },
        "PlotApicalPoint/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: ["morph_path", "apical_point_path", "plot_apical_point_path"]
                }
            }
        },
        "CollectCurated/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: ["morph_path"],
                }
            }
        },
        "CutLeaves/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: ["morph_path", "cut_leaves_path"],
                    clean_exception(
                        "ValueError: Can not search for cut leaves for a neuron with no neurites."
                    ): ["exception"],
                },
            }
        },
        "PlotCutLeaves/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: ["morph_path", "cut_leaves_path", "plot_cut_leaves_path"]
                }
            }
        },
        "HardLimit/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {out_dir_pattern: ["morph_path", "hard_limit_path"]}
            }
        },
        "PlotHardLimit/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "hard_limit_path",
                        "morph_path",
                        "plot_hard_limit_path",
                    ],
                    clean_exception(
                        "ValueError: Can not plot axis marker for a neuron with no neurites."
                    ): ["exception"],
                }
            }
        },
    }

    # Check the results
    assert_equal_trees(
        annotation_expected_dir, example_1 / "out_annotated", specific_args=annotated_specific_args
    )

    # Run the Repair workflow
    assert luigi.build([workflows.Repair()], local_scheduler=True)

    # Define post-processes to check the results
    repair_expected_dir = data_dir / "test_example_1" / "out_repaired"
    repaired_specific_args = {
        "repaired_dataset.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "apical_point_path",
                        "cut_leaves_path",
                        "error_annotated_path",
                        "error_marker_path",
                        "hard_limit_path",
                        "marker_path",
                        "morph_path",
                        "repair_morph_db_path_asc",
                        "repair_morph_db_path_h5",
                        "repair_morph_db_path_swc",
                        "repair_morph_path",
                        "unravel_morph_db_path_asc",
                        "unravel_morph_db_path_h5",
                        "unravel_morph_db_path_swc",
                        "unravel_morph_path",
                        "zero_diameter_morph_db_path_asc",
                        "zero_diameter_morph_db_path_h5",
                        "zero_diameter_morph_db_path_swc",
                        "zero_diameter_morph_path",
                        "smooth_morph_path",
                    ]
                }
            }
        },
        "Repair/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "('CollectAnnotated', 'apical_point_path')",
                        "('CollectAnnotated', 'cut_leaves_path')",
                        "('CollectAnnotated', 'morph_path')",
                        "('FixZeroDiameters', 'morph_path')",
                        "('MakeRelease', 'repair_morph_db_path_asc')",
                        "('MakeRelease', 'repair_morph_db_path_h5')",
                        "('MakeRelease', 'repair_morph_db_path_swc')",
                        "('MakeRelease', 'repair_morph_path')",
                        "('MakeRelease', 'unravel_morph_db_path_asc')",
                        "('MakeRelease', 'unravel_morph_db_path_h5')",
                        "('MakeRelease', 'unravel_morph_db_path_swc')",
                        "('MakeRelease', 'unravel_morph_path')",
                        "('MakeRelease', 'zero_diameter_morph_db_path_asc')",
                        "('MakeRelease', 'zero_diameter_morph_db_path_h5')",
                        "('MakeRelease', 'zero_diameter_morph_db_path_swc')",
                        "('MakeRelease', 'zero_diameter_morph_path')",
                        "('PlotRepair', 'cut_leaves_path')",
                        "('PlotRepair', 'morph_path')",
                        "('PlotRepair', 'plot_repair_path')",
                        "('RepairNeurites', 'morph_path')",
                        "('RepairNeurites', 'unravelled_apical_point_path')",
                        "('RepairNeurites', 'unravelled_cut_leaves_path')",
                        "('Unravel', 'apical_point_path')",
                        "('Unravel', 'cut_leaves_path')",
                        "('Unravel', 'morph_path')",
                        "('Unravel', 'unravelled_apical_point_path')",
                        "('Unravel', 'unravelled_cut_leaves_path')",
                        "('SmoothDiameters', 'morph_path')",
                        "('SmoothDiameters', 'apical_point_path')",
                        "('PlotSmoothDiameters', 'morph_path')",
                        "('PlotSmoothDiameters', 'smooth_morph_path')",
                        "('PlotSmoothDiameters', 'plot_smooth_path')",
                        "('MakeCollage', 'morph_path')",
                        "apical_point_path",
                        "cut_leaves_path",
                        "morph_path",
                        "repair_morph_db_path_asc",
                        "repair_morph_db_path_h5",
                        "repair_morph_db_path_swc",
                        "repair_morph_path",
                        "unravel_morph_db_path_asc",
                        "unravel_morph_db_path_h5",
                        "unravel_morph_db_path_swc",
                        "unravel_morph_path",
                        "zero_diameter_morph_db_path_asc",
                        "zero_diameter_morph_db_path_h5",
                        "zero_diameter_morph_db_path_swc",
                        "zero_diameter_morph_path",
                        "smooth_morph_path",
                    ],
                },
            }
        },
        "CollectAnnotated/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "apical_point_path",
                        "cut_leaves_path",
                        "morph_path",
                    ],
                }
            }
        },
        "RepairNeurites/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "morph_path",
                        "unravelled_apical_point_path",
                        "unravelled_cut_leaves_path",
                    ],
                },
            }
        },
        "FixZeroDiameters/report.csv": {
            "format_data_kwargs": {"replace_pattern": {out_dir_pattern: ["morph_path"]}}
        },
        "MakeRelease/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "repair_morph_db_path_asc",
                        "repair_morph_db_path_h5",
                        "repair_morph_db_path_swc",
                        "repair_morph_path",
                        "unravel_morph_db_path_asc",
                        "unravel_morph_db_path_h5",
                        "unravel_morph_db_path_swc",
                        "unravel_morph_path",
                        "zero_diameter_morph_db_path_asc",
                        "zero_diameter_morph_db_path_h5",
                        "zero_diameter_morph_db_path_swc",
                        "zero_diameter_morph_path",
                    ]
                }
            }
        },
        "Unravel/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: [
                        "apical_point_path",
                        "cut_leaves_path",
                        "morph_path",
                        "unravelled_apical_point_path",
                        "unravelled_cut_leaves_path",
                    ]
                }
            }
        },
        "PlotRepair/report.csv": {
            "format_data_kwargs": {
                "replace_pattern": {
                    out_dir_pattern: ["morph_path", "cut_leaves_path", "plot_repair_path"]
                }
            }
        },
        "MakeCollage/report.csv": {
            "format_data_kwargs": {"replace_pattern": {out_dir_pattern: ["morph_path"]}}
        },
    }

    # Check the results
    assert_equal_trees(
        repair_expected_dir, example_1 / "out_repaired", specific_args=repaired_specific_args
    )



"""Tests for the morphology_workflows.cli module."""
import re

from morphology_workflows.tasks import cli


class TestCLI:
    """Test the CLI of the morphology-workflows package."""

    def test_help(self, capsys):
        """Test the --help argument."""
        try:
            cli.main(arguments=["--help"])
        except SystemExit:
            pass
        captured = capsys.readouterr()
        assert (
            re.match(
                r"usage: \S+ .*Run the workflow\n\npositional arguments:\s*"
                r"{Fetch,Placeholders,Curate,Annotate,Repair}\s*Possible workflows.*",
                captured.out,
                flags=re.DOTALL,
            )
            is not None
        )

    def test_dependency_graph(self, tmpdir, data_dir):
        """Test the --help argument."""
        output_path = (data_dir / "neuromorpho_config_download.json").resolve()
        cli.main(
            arguments=[
                "--create-dependency-graph",
                str(tmpdir / "dependency_graph.png"),
                "Fetch",
                "--source",
                "NeuroMorpho",
                "--config-file",
                str(output_path),
            ]
        )

        assert output_path.exists()



"""Placeholders tasks."""
import json
import logging

from data_validation_framework.target import TaggedOutputLocalTarget
from data_validation_framework.task import TagResultOutputMixin
from luigi.parameter import OptionalPathParameter
from luigi.parameter import OptionalStrParameter
from luigi.parameter import PathParameter
from luigi_tools.task import WorkflowTask

from morphology_workflows.placeholders import compute_placeholders
from morphology_workflows.tasks.fetch import Fetch

logger = logging.getLogger(__name__)


class Placeholders(TagResultOutputMixin, WorkflowTask):
    """Compute the place holders for a given region and mtype set."""

    region = OptionalStrParameter(
        description=":str: The region to consider.",
        default=None,
    )
    mtype = OptionalStrParameter(
        description=":str: The mtype to consider.",
        default=None,
    )
    config = OptionalPathParameter(
        description=":str: The path to the JSON config file.",
        exists=True,
        default=None,
    )
    input_dir = OptionalPathParameter(
        description=":str: The directory containing the morphologies and their metadata file.",
        exists=True,
        default=None,
    )
    result_path = PathParameter(
        default="placeholders.csv", description=":str: Path to the output file."
    )

    def requires(self):
        if self.input_dir is None:  # pragma: no cover
            return Fetch()
        return None

    def run(self):
        input_dir = self.input_dir or self.input()["morphologies"]
        if self.config is not None:
            with self.config.open() as f:  # pylint: disable=no-member
                config = json.load(f)
        else:
            config = None
        df = compute_placeholders(input_dir, self.region, self.mtype, config)
        df.to_csv(self.output().path, index=False)

    def output(self):
        return TaggedOutputLocalTarget(
            self.result_path,
            create_parent=True,
        )
